#ifndef BLD_CHECKOP_H
#define BLD_CHECKOP_H

/******************************************************************************

Copyright (c) 1997, Andrew J. Weber, Jason P. Weber

Copyright information should have been included with this distribution
and may be found at ../meta/COPYRIGHT or obtained via email from weber@acm.org

******************************************************************************/

#include <gpl/gpl.h>
#include <bld/op.h>

/***********************************************************************
class BLD_BuildOperation

Build file modification times.
***********************************************************************/
class BLD_BuildOperation : public BLD_ListOperation
{
	public:
		static const GPL_String current;
		static const GPL_String built;
		static const GPL_String failed;

		BLD_BuildOperation(void) {}
		BLD_BuildOperation(const GPL_Array< GPL_Pointer<BLD_Operation> > & o)
			: BLD_ListOperation(o) {}
		virtual ~BLD_BuildOperation(void) {}

		virtual BLD_Value Perform(void) const;
		virtual GPL_String ToString(void) const;

		static void Update(const GPL_String & result, GPL_String & status);
		static void Build(const GPL_String & name, GPL_String & result);
};

/***********************************************************************
class BLD_TargetCheckOperation

Check target build times.
***********************************************************************/
class BLD_TargetCheckOperation : public BLD_ListOperation
{
	public:
		BLD_TargetCheckOperation(void) : descend(TRUE) {}
		BLD_TargetCheckOperation(const GPL_Array< GPL_Pointer<BLD_Operation> > & o,
			const GPL_Boolean d = TRUE) : BLD_ListOperation(o), descend(d) {}
		virtual ~BLD_TargetCheckOperation(void) {}

		inline GPL_Boolean ShouldDescend(void) const { return descend; }
		virtual BLD_Value Perform(void) const;
		virtual GPL_String ToString(void) const;

		static void Check(const GPL_String & name, const GPL_Boolean descend,
							GPL_String & result);

	private:
		GPL_Boolean descend;
};

/***********************************************************************
class BLD_FileCheckOperation

Check file modification times.
***********************************************************************/
class BLD_FileCheckOperation : public BLD_ListOperation
{
	public:
		BLD_FileCheckOperation(void) {}
		BLD_FileCheckOperation(const GPL_Array< GPL_Pointer<BLD_Operation> > & o)
			: BLD_ListOperation(o) {}
		virtual ~BLD_FileCheckOperation(void) {}

		virtual BLD_Value Perform(void) const;
		virtual GPL_String ToString(void) const;

		static GPL_Boolean IsNew(const GPL_String & name);
};

/***********************************************************************
class BLD_CheckOperation

Check target and file modification times.
***********************************************************************/
class BLD_CheckOperation : public BLD_ListOperation
{
	public:
		BLD_CheckOperation(void) {}
		BLD_CheckOperation(const GPL_Array< GPL_Pointer<BLD_Operation> > & o)
			: BLD_ListOperation(o) {}
		virtual ~BLD_CheckOperation(void) {}

		virtual BLD_Value Perform(void) const;
		virtual GPL_String ToString(void) const;
};

#endif /* BLD_CHECKOP_H */
