/******************************************************************************

COPYRIGHT_HEADER

******************************************************************************/

#include <iws/filewin.h>

/***********************************************************************
class IWS_FileSelector
***********************************************************************/
IWS_FileSelector::IWS_FileSelector(void)
{
	fileSelectHandler = new IWS_FileSelectHandler(this);
	fileUpdateDirectoryHandler = new IWS_FileUpdateDirectoryHandler(this);
	fileUpdateFileHandler = new IWS_FileUpdateFileHandler(this);
	fileUpdateFilterHandler = new IWS_FileUpdateFilterHandler(this);
	applyHandler = NULL;

	SetDirection(WDS_VERTICAL);
	SetFlags(WDS_PARTITION_FLAGS_ATTACH);
	verticalForm = new WDS_Form();
	verticalForm->SetMode(WDS_FORM_SINGLE_COLUMN);
	verticalForm->SetSizing(EW_HORIZONTAL, WDS_SIZING_MANUAL);
	verticalForm->SetSizing(EW_VERTICAL, WDS_SIZING_FIT_CHILDREN);

	dirField = new IWS_GetString();
	dirField->SetName("dirField");
	dirField->SetAnnotation("Directory:");
	dirField->SetHandler(fileUpdateDirectoryHandler);

	fileField = new IWS_GetString();
	fileField->SetName("fileField");
	fileField->SetAnnotation("Selection:");
	fileField->SetHandler(fileUpdateFileHandler);

	filterField = new IWS_GetString();
	filterField->SetName("filterField");
	filterField->SetAnnotation("Filter:");
	filterField->SetHandler(fileUpdateFilterHandler);

	fileList = new IWS_StringList();
	fileList->SetName("fileList");
	fileList->SetHandler(fileSelectHandler);

	verticalForm->AddWidget(dirField);
	verticalForm->AddWidget(fileField);
	verticalForm->AddWidget(filterField);
	AddWidget(verticalForm);
	AddWidget(fileList);

	AlignAnnotations();

	GPL_String p(256, 0);
	osdCWDGet(p.GetData(), p.GetSize());
	SetDirectory(p);
	SetFilter(&fileFilter);
	SetFilterString("*");

	WDS_String *s = fileList->GetTextList()->GetStringList()->ToIndex(0);
	if ((WDS_String *) 0 != s)
	{
		fileField->SetText(s->Buffer());
	}
}

IWS_FileSelector::~IWS_FileSelector(void)
{
	delete fileSelectHandler;
	delete fileUpdateDirectoryHandler;
	delete fileUpdateFileHandler;
	delete fileUpdateFilterHandler;
}

void IWS_FileSelector::AlignAnnotations(void)
{
#if 0
	long w,l;
	w=dirField->GetLocalLook()->
		GetElement(WDS_STATE_ALL,WDS_ELEMENT_ANNOTATION_X);
	l=filterField->GetLocalLook()->
		GetElement(WDS_STATE_ALL,WDS_ELEMENT_ANNOTATION_X);
	if (l>w)w=l;
	l=fileField->GetLocalLook()->
		GetElement(WDS_STATE_ALL,WDS_ELEMENT_ANNOTATION_X);
	if (l>w)w=l;
	l=dirField->GetLocalLook()->
		GetElement(WDS_STATE_ALL,WDS_ELEMENT_ANNOTATION_Y);
	dirField->SetAnnotationSize(w,l);
	filterField->SetAnnotationSize(w,l);
	fileField->SetAnnotationSize(w,l);
#else
	IWS_AnnotationAlignableGroup group;
	group.AppendNode(dirField);
	group.AppendNode(filterField);
	group.AppendNode(fileField);
	group.Align();
#endif
}

void IWS_FileSelector::UpdateFileList(void)
{
	// Update the text list
	GPL_List<GPL_String> files;
	fileList->Clear();
	fileList->AddText("..");
	if (GPL_FileSelector::GetFiles(files))
	{
		for (register unsigned long i = 0; files.GetCount() > i; ++i)
		{
			OSD_FileAttributes attr;
			GPL_String s;
			s = dirField->GetText();
			if (s[s.GetLength()-1] != OSD_DIRECTORY_SEPARATOR)
			{
				s.Append(OSD_DIRECTORY_SEPARATOR);
			}
			s.Append(files[i].GetData());
			osdFileGetAttributes((char *)s, &attr);
			if (attr.attr & OSD_FILE_ATTR_DIR)
			{
				fileList->AddText(files[i].GetData(), WDS_STRING_STYLE_BOLD);
			}	
			else
			{
				fileList->AddText(files[i].GetData());
			}
		}
	}
}

void IWS_FileSelector::Apply(WDS_Activation *actv, const GPL_String &string)
{
	if (applyHandler)
	{
		applyHandler->SetActivation(actv);
		applyHandler->SetFileSelector(this);
		applyHandler->HandleEvent(string);
	}
}

void IWS_FileSelector::SetDirectory(const GPL_String & d)
{
	// Save the directory.
	directory = GPL_File::GetFullPathName(d.GetData());
	
	// Update the text field.
	dirField->SetText(directory.GetData());

	// Execute the filter.
	UpdateFileList();
}

void IWS_FileSelector::SetFilterString(const GPL_String & f)
{
	// Save the pattern.
	fileFilter.SetPattern(f);

	// Update the text field.
	filterField->SetText(f.GetData());

	// Execute the filter.
	UpdateFileList();
}

void IWS_FileSelector::SetFile(const GPL_String & f)
{
	// Save the file name.
	file = f;

	// Update the text field.
	fileField->SetText(file.GetData());
}

const GPL_String & IWS_FileSelector::GetSelectedFile(void)
{
	selectedFile.Clear();
	selectedFile.Append(dirField->GetText());
	if (selectedFile[selectedFile.GetLength()-1] != OSD_DIRECTORY_SEPARATOR)
	{
		selectedFile.Append(OSD_DIRECTORY_SEPARATOR);
	}
	selectedFile.Append(fileField->GetText());
	
	return selectedFile;
}

void IWS_FileSelector::SetFilter(GPL_FileFilter *f)
{
	// Save the filter.
	filter = f;

	// Execute the filter.
	UpdateFileList();
}

/***********************************************************************
class IWS_FileSelectHandler
***********************************************************************/
void IWS_FileSelectHandler::HandleEvent(const GPL_String &string)
{
	EW_Event *event = activation->GetEvent();

	char *str;
	str = fileSelector->GetFileList()->GetSelectedString();
	if (str)
	{
		if (!strcmp(str, "..") && event && (event->GetRepetition() == 1))
		{
			GPL_String s;
			s = fileSelector->GetDirField()->GetText();
			while (s.GetLength() > 1 &&
				(s[s.GetLength()-1] == OSD_DIRECTORY_SEPARATOR))
			{
				s.Remove(s.GetLength()-1, 1);
			}
			while (s.GetLength() && 
				(s[s.GetLength()-1] != OSD_DIRECTORY_SEPARATOR))
			{
				s.Remove(s.GetLength()-1, 1);
			}
			fileSelector->SetDirectory(s);
		}
		else
		{
			OSD_FileAttributes attr;
			GPL_String s;
			s = fileSelector->GetDirectory();
			if (s[s.GetLength()-1] != OSD_DIRECTORY_SEPARATOR)
			{
				s.Append(OSD_DIRECTORY_SEPARATOR);
			}
			s.Append(str);
			osdFileGetAttributes((char *)s, &attr);
			if (attr.attr & OSD_FILE_ATTR_DIR)
			{
				if (event && (event->GetRepetition() == 1))
				{
					fileSelector->SetDirectory(s);
				}
			}
			else
			{
				fileSelector->SetFile(str);
				if (event && (event->GetRepetition() == 1))
				{
					fileSelector->Apply(NULL, "DOUBLECLICK");
				}
			}
		}
		osdFree(str);
	}
}

/***********************************************************************
class IWS_FileUpdateDirectoryHandler
***********************************************************************/
void IWS_FileUpdateDirectoryHandler::HandleEvent(const GPL_String &string)
{
	fileSelector->SetDirectory(fileSelector->GetDirField()->GetText());
}

/***********************************************************************
class IWS_FileUpdateFileHandler
***********************************************************************/
void IWS_FileUpdateFileHandler::HandleEvent(const GPL_String &string)
{
	fileSelector->SetFile(fileSelector->GetFileField()->GetText());
}

/***********************************************************************
class IWS_FileUpdateFilterHandler
***********************************************************************/
void IWS_FileUpdateFilterHandler::HandleEvent(const GPL_String &string)
{
	fileSelector->SetFilterString(fileSelector->GetFilterField()->GetText());
}

/***********************************************************************
class IWS_FileSelectorWindowButtonHandler
***********************************************************************/
void IWS_FileSelectorWindowButtonHandler::HandleEvent(const GPL_String &string)
{
	if (string == "Apply")
	{
		OSD_FileAttributes attr;
		osdFileGetAttributes(window->GetSelectedFile().GetData(), &attr);
		if (OSD_FILE_ATTR_DIR != (attr.attr & OSD_FILE_ATTR_DIR))
		{
			window->GetFileSelector()->Apply(activation, "APPLY");
			window->Hide();
		}
	}
	else if (string == "Cancel")
	{
		window->Hide();
	}
}

/***********************************************************************
class IWS_FileSelectorWindow
***********************************************************************/
IWS_FileSelectorWindow::IWS_FileSelectorWindow(void)
{
	fileSelector = new IWS_FileSelector();
	buttonRow = new IWS_ButtonRow();
	handler = new IWS_FileSelectorWindowButtonHandler(this);
	relay = new IWS_FileApplyRelayHandler(this);

	buttonRow->AddButton("Apply", handler); 
	buttonRow->AddButton("Cancel", handler); 
	fileSelector->SetHandler(relay);

	IncludeWidget(fileSelector);
	IncludeWidget(buttonRow);
}

IWS_FileSelectorWindow::~IWS_FileSelectorWindow(void)
{
	delete fileSelector;
	delete buttonRow;
	delete handler;
	delete relay;
}

void IWS_FileSelectorWindow::SetApplyHandler(IWS_FileApplyHandler *h)
{
	relay->SetHandler(h);
}

/***********************************************************************
class IWS_FileApplyRelayHandler
***********************************************************************/
void IWS_FileApplyRelayHandler::HandleEvent(const GPL_String &string)
{
	if ((IWS_Handler *) GPL_BAD_ADDRESS != applyHandler)
	{
		applyHandler->SetFileSelector(GetFileSelector());
		applyHandler->SetActivation(activation);
		applyHandler->HandleEvent(string);
	}
	window->Hide();
}
