
/******************************************************************************

COPYRIGHT_HEADER

******************************************************************************/

#include <iws/iws.h>

IWS_DefaultGetStringHandler		IWS_GetString::defaultHandler;
IWS_DefaultStringListHandler	IWS_StringList::defaultHandler;

IWS_GetString::IWS_GetString(const GPL_String &n, const long tw, 
								const long th, const long aw, const long ah)
{
	IWS_GetStringConstructor(tw,th,aw,ah);
	GPL_String s = n;
	SetName((char *)s);
	SetAnnotation((const char *)s);
}

IWS_GetString::IWS_GetString(const long tw, const long th,
								const long aw, const long ah)
{
	IWS_GetStringConstructor(tw,th,aw,ah);
}

void IWS_GetString::IWS_GetStringConstructor(const long tw, const long th,
								const long aw, const long ah)
{
	SetGeneralFlags(WDS_WIDGET_FLAGS_AUTOSIZE_Y);
	if(-1 == tw)
	{
		SetBounds(EW_HORIZONTAL, EW_BOUNDS_NONE);
	}
	SetAnnotationMode(WDS_WIDGET_ANNOTATION_LEFT);
	SetTextSize(tw, th);
	SetAnnotationSize(aw, ah);
	SetHandler(&defaultHandler);
}

IWS_GetString::~IWS_GetString()
{
}

const char *IWS_GetString::GetText(void)
{
	return ((IWS_GetString *) this)->GetString()->Buffer();
}

void IWS_GetString::SetText(const char *text)
{
	SetString((char *) text);
	UpdateSize();
}

IWS_AnnotationAlignable_IMPLEMENTATION(IWS_GetString, WDS_GetString);

void IWS_GetString::SetHandler(IWS_Handler *handler, WDS_Activation *actv)
{
	if(!handler) return;
	WDS_Activation a;
	if(!actv) actv = &a;
	actv->SetObject((void *)handler);
	SetStateChangeCallback(WDS_STATE_0, handler->Callback, actv);
}

IWS_StringList::IWS_StringList()
{
	IWS_StringListConstructor();
}

IWS_StringList::IWS_StringList(const GPL_String &n)
{
	IWS_StringListConstructor();
	GPL_String s = n;
	SetName((char *)s);
}

void IWS_StringList::IWS_StringListConstructor(void)
{
	SetSizing(EW_HORIZONTAL,WDS_SIZING_MANUAL);
	SetSizing(EW_VERTICAL,WDS_SIZING_MANUAL);
	SetMode(WDS_FORM_COLUMN);

	scrollRegion = new WDS_ScrollRegion();
	scrollRegion->GetLocalLook()->SetElement(WDS_STATE_0,WDS_ELEMENT_SCROLLREGION_SCROLLABLE,WDS_SCROLLABLE_OFF);

	textList = new WDS_TextList();
	textList->SetBounds(EW_HORIZONTAL,EW_BOUNDS_FILL_PARENT);
	textList->SetEditable(WDS_TEXTLIST_EDITABLE_NONE);

	scrollRegion->AddWidget(textList);
	AddWidget(scrollRegion);

	strings = new WDS_StringList();
	textList->SetStringList(strings);
	UpdateSize();

	SetHandler(&defaultHandler);
}

IWS_StringList::~IWS_StringList(void)
{
	Clear();
}

char *IWS_StringList::GetSelectedString()
{
	return textList->GetSelectedString();
}

void IWS_StringList::AddText(const char *text, long style)
{
	register WDS_String *string = new WDS_String();
	string->SetStyle(style);
	string->NewBuffer((char *) text);
	strings->AppendNode(string);
	UpdateSize();
}

void IWS_StringList::Clear(void)
{
	strings->ToHead();
	register WDS_String *string;
	while ((WDS_String *) GPL_BAD_ADDRESS != (string = strings->PostIncrement()))
	{
		delete string;
	}
	strings->Clear();
	textList->SetDirty(EW_DIRTY_COMPLETE);
	UpdateSize();
}

void IWS_StringList::UpdateSize(void)
{
	long w = 0;
	long h = 0;
	long tw = -1;
	long th = -1;
	register WDS_String *string;
	strings->ToHead();
	while ((WDS_String *) GPL_BAD_ADDRESS != (string = strings->PostIncrement()))
	{
		IWS_APP->GetApproximateSize(string->Buffer(), tw, th);
		if (tw > w)
		{
			w = tw;
		}
		h += th;
	}
	textList->SetPrefSize(w, h);
}

void IWS_StringList::Dump(void)
{
	register unsigned long i = 0;
	register WDS_String *string;
	strings->ToHead();
	while ((WDS_String *) GPL_BAD_ADDRESS != (string = strings->PostIncrement()))
	{
		printf("%d: %s\n", i++, string->Buffer());
	}
}

void IWS_StringList::SetHandler(IWS_Handler *handler, WDS_Activation *actv)
{
	if(!handler) return;
	WDS_Activation a;
	if(!actv) actv = &a;
	actv->SetObject((void *)handler);
	textList->SetStateChangeCallback(WDS_STATE_2, handler->Callback, actv);
}

