/******************************************************************************

Coral Tree iws/statbtn.cc
Copyright (C) 1998 Infinite Monkeys Incorporated

This program is free software; you can redistribute it and/or modify
it under the terms of the Artistic License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
Artistic License for more details.

You should have received a copy of the Artistic License along with
this program (see meta/ART_LICE); if not, write to coral@imonk.com.

******************************************************************************/

#include <iws/statbtn.h>
#include <iws/app.h>

/***********************************************************************
class IWS_StatusViewerWindow : public IWS_VerticalWindow
***********************************************************************/
IWS_StatusViewerWindow::IWS_StatusViewerWindow(void)
{
	stata = new IWS_StringList();

	hideHandler = new IWS_HideWindowHandler((IWS_Window *) this);

	buttonBar = new IWS_ButtonBar();
	buttonBar->AddButton("Close", hideHandler);

	GetRootPartition()->AddWidget(stata);
	GetRootPartition()->AddWidget(buttonBar);

	long w, h;
	IWS_APP->GetApproximateSize("X", w, h);
	h = (((float)h)*0.7);
	EW_FontRequest request;
	request.SetHeight(h);
	request.SetStyle(EW_FONT_FIXED);
	EW_Font font;
	if (!IWS_APP->GetEventContext()->GetPhantom()->LoadFont(&font, &request))
	{
		SetFont(&font);
	}
}

IWS_StatusViewerWindow::~IWS_StatusViewerWindow(void)
{
	delete hideHandler;
}

void IWS_StatusViewerWindow::SetStata(const GPL_List<GPL_String> & s)
{
	stata->Clear();
	for (unsigned long i = 0; s.GetCount() > i; ++i)
	{
		stata->AddText(s[i]);
	}
}

/***********************************************************************
class IWS_StatusButton : public IWS_HorizontalForm
***********************************************************************/
IWS_StatusButton::IWS_StatusButton(void) : history(50)
{
	window = new IWS_StatusViewerWindow();

	handler = new IWS_ShowStatusViewerWindowHandler(this, window);

	SetGeneralFlags(WDS_WIDGET_FLAGS_NONE);
//	SetAnnotationMode(WDS_WIDGET_ANNOTATION_LEFT);
	long width = 0;
	long height = 0;
//	IWS_APP->GetApproximateSize("Status", width, height);
//	SetAnnotationSize(width, height);
//	SetAnnotation("Status");
	GPL_String initial(80, ' ');
	IWS_APP->GetApproximateSize(initial, width, height);
	SetTextSize(width, height);
	SetLabel(initial);
	SetHandler(handler);
	attachedToLogger = FALSE;
}

IWS_StatusButton::~IWS_StatusButton(void)
{
	if(attachedToLogger)
	{
		WBA_LOGGER.SetDelivery("BUILTIN_WARNING",
			new GPL_DeliveryOsdOutput());
		WBA_LOGGER.SetDelivery("BUILTIN_ERROR", 
			new GPL_DeliveryOsdOutputTerse());
	}
	delete handler;
	delete window;
}

void IWS_StatusButton::SetStatus(const GPL_String & s)
{
	SetLabel(s.GetData());
	SetDirty(EW_DIRTY_COMPLETE);

	// Enforce the history limit.
	while (history <= stata.GetCount())
	{
		stata.Remove(0);
		if(window->GetMapped())
		{
			WDS_String *s;
			s = window->GetStata()->GetTextList()->GetStringList()->ToHead();
			if(s)
			{
				window->GetStata()->GetTextList()->GetStringList()->RemoveNode(s);
			}
		}
	}
	stata.Append(s);
	SetDirty(EW_DIRTY_COMPLETE);
	if(window->GetMapped())
	{
		window->GetStata()->AddText(s);
		window->GetStata()->SetDirty(EW_DIRTY_COMPLETE);
	}
}

void IWS_StatusButton::AttachLogger(void)
{
	WBA_LOGGER.SetDelivery("BUILTIN_WARNING", 
		new IWS_StatusButtonDelivery(this));
	WBA_LOGGER.SetDelivery("BUILTIN_ERROR", 
		new IWS_StatusButtonDelivery(this));
	attachedToLogger = TRUE;
}

GPL_Boolean IWS_StatusButtonDelivery::Deliver(	const GPL_String & filename,
												const long linenum,
												const GPL_String & tname,
												const GPL_String & text)
{
	statusButton->SetStatus(TerseImage(tname, text));
	return TRUE;
}

/***********************************************************************
class IWS_ShowStatusViewerWindowHandler
***********************************************************************/
void IWS_ShowStatusViewerWindowHandler::HandleEvent(const GPL_String &string)
{
	if ((IWS_StatusButton *) GPL_BAD_ADDRESS == button ||
		(IWS_StatusViewerWindow *) GPL_BAD_ADDRESS == window)
	{
		return;
	}

	window->SetStata(button->GetStata());
	window->Show();
}
