#ifndef IWS_WINDOW_H
#define IWS_WINDOW_H

/******************************************************************************

Coral Tree iws/window.h
Copyright (C) 1998 Infinite Monkeys Incorporated

This program is free software; you can redistribute it and/or modify
it under the terms of the Artistic License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
Artistic License for more details.

You should have received a copy of the Artistic License along with
this program (see meta/ART_LICE); if not, write to coral@imonk.com.

******************************************************************************/

#include <wds/wds.h>
#include <iws/handler.h>

class IWS_Window : public EW_Window
{
	public:
		WDS_Scope		*rootScope;
	public:
		IWS_Window(const char *title = (const char *) GPL_BAD_ADDRESS);
virtual	~IWS_Window(void);

	/* CHILD RELATED */
virtual	void IncludeWidget(WDS_Widget *widget);

	/* SELF ORIENTED */
virtual GPL_Boolean IsShown (void) const;

virtual void Show (void);
virtual void Hide (void);

	/* GENERAL WINDOW FUNCTIONALITY */
static GPL_Boolean IsWindowShown (EW_Window * window);
static void ShowWindow (EW_Window * window);
static void HideWindow (EW_Window * window);
};

class IWS_VerticalWindow : public IWS_Window
{
	protected:
		WDS_Partition	*rootPartition;
	public:
		IWS_VerticalWindow(const char *title = (const char *) GPL_BAD_ADDRESS);
virtual	~IWS_VerticalWindow(void);

	/* CHILD RELATED */
virtual	void IncludeWidget(WDS_Widget *widget);

	WDS_Partition *GetRootPartition(void){ return rootPartition; }
		
};

class IWS_HorizontalWindow : public IWS_Window
{
	protected:
		WDS_Partition	*rootPartition;
	public:
		IWS_HorizontalWindow(void);
virtual	~IWS_HorizontalWindow(void);

	/* CHILD RELATED */
virtual	void IncludeWidget(WDS_Widget *widget);

	WDS_Partition *GetRootPartition(void){ return rootPartition; }
		
};

class IWS_ShowWindowHandler : public IWS_Handler
{
	public:
		IWS_ShowWindowHandler(IWS_Window *w = (IWS_Window *) GPL_BAD_ADDRESS)
			: window(w) {}
		virtual ~IWS_ShowWindowHandler(void) {}

		virtual void HandleEvent(const GPL_String &string);

	protected:
		IWS_Window *window;
};

class IWS_HideWindowHandler : public IWS_Handler
{
	public:
		IWS_HideWindowHandler(IWS_Window *w = (IWS_Window *) GPL_BAD_ADDRESS)
			: window(w) {}
		virtual ~IWS_HideWindowHandler(void) {}

		virtual void HandleEvent(const GPL_String &string);

	protected:
		IWS_Window *window;
};

#endif /* IWS_WINDOW_H */
