
/******************************************************************************

COPYRIGHT_HEADER

******************************************************************************/

#include <mte/mte.h>

MTE_Mutex MTE_MemoryEnvironment::memory_protect;
static MTE_MemoryEnvironment MTE_MEMENV;
#define MTE_MEMORY_MUTEX (MTE_MEMENV.memory_protect)

#ifndef MTE_STANDALONE
/******************************************************************************
REPLACING GLOBAL NEW AND DELETE
This is only really necessary if there is 'real' threading going on. 
The global new calls osdMalloc, not mteMalloc directly just in case someone
wants to replace osdMalloc with yet another implementation.
Same goes for delete and free.
******************************************************************************/
#ifndef MTE_NO_GLOBAL_NEW
#ifdef MTE_IMPL
void * operator new(size_t size)
{
	return osdMalloc(size);
}
void operator delete(void *ptr)
{
	if(ptr)
	{
		osdFree(ptr);
	}
}
#if OSD_OS!=OSD_IRIX && OSD_OS!=OSD_SUNOS && OSD_COMPILER!=OSD_ANSI && OSD_COMPILER!=OSD_VISUAL 
void operator delete(void *ptr, size_t size)
{
	if(ptr)
	{
		osdFree(ptr);
	}
}
#endif
#endif
#endif

void mteInstallAllocPtrs()
{
/******************************************************************************
These replacements are only necessary if there is 'real' threading going on.
******************************************************************************/
#ifdef MTE_IMPL
	osdMallocImplPtr = &mteMalloc;
	osdCallocImplPtr = &mteCalloc;
	osdReallocImplPtr = &mteRealloc;
	osdFreeImplPtr = &mteFree;
	osdMemsetImplPtr = &mteMemset;
	osdMemcpyImplPtr = &mteMemcpy;
	osdMemcmpImplPtr = &mteMemcmp;
	osdMemmoveImplPtr = &mteMemmove;
#endif
}
#endif /* MTE_STANDALONE */

void *mteMalloc(size_t x)
{
	MTE_Guard(&MTE_MEMORY_MUTEX);
	return osdMallocImpl(x);
}

void *mteCalloc(size_t x, size_t y)
{
	MTE_Guard(&MTE_MEMORY_MUTEX);
	return osdCallocImpl(x,y);
}

void  mteFree(void *x)
{
	MTE_Guard(&MTE_MEMORY_MUTEX);
	osdFreeImpl(x);
}

void *mteMemset(void *x, int y, size_t z)
{
	MTE_Guard(&MTE_MEMORY_MUTEX);
	return osdMemsetImpl(x,y,z);
}

void *mteMemcpy(void *x, const void *y, size_t z)
{
	MTE_Guard(&MTE_MEMORY_MUTEX);
	return osdMemcpyImpl(x,y,z);
}

int mteMemcmp(const void *x, const void *y, size_t z)
{
	MTE_Guard(&MTE_MEMORY_MUTEX);
	return osdMemcmpImpl(x,y,z);
}

void *mteMemmove(void *x, const void *y, size_t z)
{
	MTE_Guard(&MTE_MEMORY_MUTEX);
	return osdMemmoveImpl(x,y,z);
}

void *mteRealloc(void *x, size_t y)
{
	MTE_Guard(&MTE_MEMORY_MUTEX);
	/* CANNOT USE OSD's IMPLEMENTATION AS THIS WILL CAUSE DEADLOCK */
	if(!x)
	{
		return osdMallocImpl(y);
	}
	else
	{
#if (OSD_OS==OSD_SUNOS && OSD_OSVER==4 && OSD_COMPILER!=OSD_GNU) && defined __cplusplus
		return (void *)realloc((char *)x,y);
#else
		return realloc(x,y);
#endif
	}
}

