
/******************************************************************************

Coral Tree nl/channel.cc
Copyright (C) 1998 Infinite Monkeys Incorporated

This program is free software; you can redistribute it and/or modify
it under the terms of the Artistic License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
Artistic License for more details.

You should have received a copy of the Artistic License along with
this program (see meta/ART_LICE); if not, write to coral@imonk.com.

******************************************************************************/

#include "nl.h"

#define NL_RETRY	3

long NL_Channel::count = 0;

/******************************************************************************
	NL_Channel::NL_Channel(void)

	create a channel object

Keywords: NL_Channel, network
******************************************************************************/
NL_Channel::NL_Channel(void)
{
	NL_PRINT(NL_NL, NL_LOG, "NL_Channel::NL_Channel");
	count++;

	last_errno = NL_CHANNEL_NOERROR;

	SetValid(TRUE);

	/* winsock will keep counter of number of times called */
	NL_IP_SOCKET_STARTUP();

	connection_type = NL_CONNECTIONLESS;
	protocol = 0;
	domain = 0;

	long m = NL_RETRY;
	while (0 < (m--) && !(inet_address[NL_LOCAL] = nlGetHostAddress()))
	{
		NL_PRINT(NL_NL, NL_WARN,
				"NL_Channel::NL_Channel retry nlGetHostAddress()");
	}

	inet_address[NL_REMOTE] = 0;
	inet_address[NL_RETURN] = 0;

	inet_port[NL_LOCAL] = 0;
	inet_port[NL_REMOTE] = 0;
	inet_port[NL_RETURN] = 0;

	ip_handle = INVALID_SOCKET_HANDLE;

	open = FALSE;
	q_specified = FALSE;

	blocking = TRUE;

	if (0 == inet_address[NL_LOCAL])
	{
		SetValid(FALSE);
	}
}

/******************************************************************************
	NL_Channel::~NL_Channel(void)

	destroy a channel object

Keywords: NL_Channel, network
******************************************************************************/
NL_Channel::~NL_Channel(void)
{
	NL_PRINT(NL_NL, NL_LOG, "NL_Channel::~NL_Channel");

	count--;

	Close();

	/* winsock will only really clean up after its counter of start-ups drops to zero */
	NL_IP_SOCKET_CLEANUP(); 
}

/******************************************************************************
unsigned short	NL_Channel::InetGetPort(char *service_name)


Keywords: NL_Channel, network
******************************************************************************/
unsigned short NL_Channel::InetGetPort(char *service_name)
{
	NL_PRINT(NL_NL, NL_LOG, "NL_Channel::InetGetPort");

	uint16 a_port;
	if (protocol & NL_IP)
	{
		switch(connection_type)
		{
		case NL_CONNECTED:
			a_port = nlGetServicePort(service_name, "tcp");
			break;

		case NL_CONNECTIONLESS:
			a_port = nlGetServicePort(service_name, "udp");
			break;

		default:
			NL_PRINT(NL_APP, NL_ERROR, "connection type not set");
			break;
		}
		if (!a_port)
		{
			NL_PRINT(NL_OS, NL_WARN, "could not find port for service");
		}
		return a_port;
	}
	else
	{
		NL_PRINT(NL_APP, NL_ERROR, "invalid protocol");
		return 0;
	}
}

/******************************************************************************
long	NL_Channel::SetPort(long location, char *service_name)

	set port by name

Keywords: NL_Channel, network
******************************************************************************/
long NL_Channel::SetPort(long location, char *service_name)
{
	NL_PRINT(NL_NL, NL_LOG, "NL_Channel::SetPort(char *)");

	uint16 port=0;
	if (domain & NL_INET)
	{
		port = InetGetPort(service_name);
	}
	else
	{
		NL_PRINT(NL_APP, NL_ERROR, "invalid domain");
		return FALSE;
	}

	if (port)
	{
		return SetPort(location, port);
	}
	else
	{
		return FALSE;
	}
}

/******************************************************************************
long	NL_Channel::SetPort(long location, uint16 port)

	set port by number

Keywords: NL_Channel, network
******************************************************************************/
long NL_Channel::SetPort(long location, uint16 port)
{
	NL_PRINT(NL_NL, NL_LOG, "NL_Channel::SetPort(uint16)");

	if (NL_RETURN == location)
	{
		NL_PRINT(NL_APP, NL_ERROR, "can not alter return port");
		return FALSE;
	}
	else if (domain & NL_INET)
	{
		inet_port[location]=port;
	}
	else
	{
		NL_PRINT(NL_APP, NL_ERROR, "invalid domain");
		return FALSE;
	}

	if (open && NL_LOCAL == location)
	{
		Close();
		Open();
	}

	return TRUE;
}

/******************************************************************************
long	NL_Channel::GetPort(long location, char *service_name,
							long service_name_len) const

	get port name

Keywords: NL_Channel, network
******************************************************************************/
long NL_Channel::GetPort(long location, char *service_name,
						long service_name_len) const
{
	NL_PRINT(NL_NL, NL_LOG, "NL_Channel::GetPort(char *)");

	if (domain & NL_INET)
	{
		if (protocol & NL_IP)
		{
			return nlGetServiceName(service_name, service_name_len,
							inet_port[location],
							(connection_type == NL_CONNECTED)? "tcp": "udp");
		}
		else
		{
			NL_PRINT(NL_APP, NL_ERROR, "invalid protocol");
			return 0;
		}
	}
	else
	{
		NL_PRINT(NL_APP, NL_ERROR, "invalid domain");
		return 0;
	}
}

/******************************************************************************
uint16	NL_Channel::GetPort(long location) const

	get port number

Keywords: NL_Channel, network
******************************************************************************/
uint16 NL_Channel::GetPort(long location) const
{
	NL_PRINT(NL_NL, NL_LOG, "NL_Channel::GetPort(void)");

	if (domain & NL_INET)
	{
		if (protocol & NL_IP)
		{
			return inet_port[location];
		}
		else
		{
			NL_PRINT(NL_APP, NL_ERROR, "invalid protocol");
			return 0;
		}
	}
	else
	{
		NL_PRINT(NL_APP, NL_ERROR, "invalid domain");
		return 0;
	}
}

/******************************************************************************
long	NL_Channel::SetAddress(long location, char *machine_name)

	set address by name

Keywords: NL_Channel, network
******************************************************************************/
long NL_Channel::SetAddress(long location, char *machine_name)
{
	NL_PRINT(NL_NL, NL_LOG, "NL_Channel::SetAddress(char *)");

	uint32 address;
	if (domain & NL_INET)
	{
		address = nlGetServerAddress(machine_name);
	}
	else
	{
		NL_PRINT(NL_APP, NL_ERROR, "invalid domain");
		return FALSE;
	}

	return SetAddress(location, address);
}

/******************************************************************************
long	NL_Channel::SetAddress(long location, uint32 machine_number)

	set address by number

Keywords: NL_Channel, network
******************************************************************************/
long NL_Channel::SetAddress(long location, uint32 machine_number)
{
	NL_PRINT(NL_NL, NL_LOG, "NL_Channel::SetAddress(uint32)");

	if (NL_LOCAL == location || NL_RETURN == location)
	{
		NL_PRINT(NL_APP, NL_ERROR, "can not alter local or return address");
		return FALSE;
	}
	else if (domain & NL_INET)
	{
		return (0 != (inet_address[location]=machine_number));
	}
	else
	{
		NL_PRINT(NL_APP, NL_ERROR, "invalid domain");
		return FALSE;
	}
}

/******************************************************************************
long  NL_Channel::GetHostname(long location) const

	get address name

Keywords: NL_Channel, network
******************************************************************************/
char *NL_Channel::GetHostname(long location) const
{
	NL_PRINT(NL_NL, NL_LOG, "NL_Channel::GetHostname(char *)");

	if (domain & NL_INET)
	{
		char *machine_name;
		machine_name = (char *) osdMalloc(NL_MAX_HOSTNAME_LEN);
		if (!machine_name)
		{
			NL_PRINT(NL_OS, NL_ERROR, "out of memory");
			return NULL;
		}
		nlGetServerHostname(machine_name, NL_MAX_HOSTNAME_LEN,
							inet_address[location]);
		return machine_name;
	}
	else
	{
		NL_PRINT(NL_APP, NL_ERROR, "invalid domain");
		return NULL;
	}
}

/******************************************************************************
uint32  NL_Channel::GetAddress(long location) const

	get address number

	return number OR zero for error

Keywords: NL_Channel, network
******************************************************************************/
uint32 NL_Channel::GetAddress(long location) const
{
	NL_PRINT(NL_NL, NL_LOG, "NL_Channel::GetAddress(void)");

	if (domain & NL_INET)
	{
		return inet_address[location];
	}
	else
	{
		NL_PRINT(NL_APP, NL_ERROR, "invalid domain");
		return 0;
	}
}

/******************************************************************************
long	NL_Channel::SetBlocking(short set)

	set blocking on or off

Keywords: NL_Channel, network
******************************************************************************/
long NL_Channel::SetBlocking(short set)
{
	NL_PRINT(NL_NL, NL_LOG, "NL_Channel::SetBlocking");

	long c;
	int n;
	if (protocol & NL_IP)
	{
		blocking = set;
		if(open)
		{
			c = set ? 0 : 1;
			n = nlIoctl(ip_handle, FIONBIO, (char *) &c);
			if (0 > n)
			{
				NL_PRINT(NL_OS, NL_WARN, "ioctl failed");
				return FALSE;
			}
		}
	}
	else
	{
		NL_PRINT(NL_APP, NL_ERROR, "invalid protocol");
		return FALSE;
	}
	
	return TRUE;
}

/******************************************************************************
unsigned short	NL_Channel::FindLocalPort(void)

	find out local port (by number) for receiving messages
		from socket info

Keywords: NL_Channel, network
******************************************************************************/
unsigned short NL_Channel::FindLocalPort(void)
{
	NL_PRINT(NL_NL, NL_LOG, "NL_Channel::FindLocalPort(void)");

	if (domain & NL_INET)
	{
		if (protocol & NL_IP)
		{
			INET_SOCK_ADDR_STRUCT addr_struct;
			memset((void *) &addr_struct, 0, sizeof(addr_struct));

			int temp_size = sizeof(addr_struct);
			int n = getsockname(ip_handle, (PTR_SOCK_ADDR_STRUCT) &addr_struct,
								&temp_size);

			if (0 != n)
			{
				NL_PRINT(NL_OS, NL_WARN, "failed to get socket name");
			}

			inet_port[NL_LOCAL] = ntohs (addr_struct.sin_port);
			return ntohs(addr_struct.sin_port);
		}
		else
		{
			NL_PRINT(NL_APP, NL_ERROR, "invalid protocol");
			return 0;
		}
	}
	else
	{
		NL_PRINT(NL_APP, NL_ERROR, "invalid domain");
		return 0;
	}
}

/******************************************************************************
long	NL_Channel::Open(void)

	open a socket and bind its name

Keywords: NL_Channel, network
******************************************************************************/
long NL_Channel::Open(void)
{
	NL_PRINT(NL_NL, NL_LOG, "NL_Channel::Open");

	INET_SOCK_ADDR_STRUCT serv_addr;
	memset((void *) &serv_addr, 0, sizeof(serv_addr));

	long return_val = 0;
	if (domain & NL_INET)
	{
		if (protocol & NL_IP)
		{
			int r;
			int optflag;
			switch (connection_type)
			{
			case NL_CONNECTED:

				// socket()
				osdBeginSignalRedoLoop();
				ip_handle = socket(PF_INET, SOCK_STREAM, 0);
				osdEndSignalRedoLoop(INVALID_SOCKET_HANDLE == ip_handle);

				if (INVALID_SOCKET_HANDLE == ip_handle)
				{
					NL_PRINT(NL_OS, NL_WARN, "could not get socket handle");
					return_val = 0;
					break;
				}

				// Allow reuse of addresses.
				optflag = 1;
				setsockopt(ip_handle, SOL_SOCKET, SO_REUSEADDR,
							(char *) &optflag, sizeof(int));

				// Don't coalesce send()'s.
				optflag = 1;
				setsockopt(ip_handle, IPPROTO_TCP, TCP_NODELAY,
							(char *) &optflag, sizeof(int));

				// bind()
				serv_addr.sin_family = PF_INET;
				serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
				serv_addr.sin_port = htons(inet_port[NL_LOCAL]);

				osdBeginSignalRedoLoop();
				r = bind(ip_handle, (PTR_SOCK_ADDR_STRUCT) &serv_addr,
						sizeof(serv_addr));
				osdEndSignalRedoLoop(0 != r);

				if (0 != r)
				{
					NL_PRINT(NL_OS, NL_WARN, "could not bind");
					nlCloseSocket(ip_handle);
					return_val =  0;
					break;
				}

				FindLocalPort();

				return_val = 1;
				open = TRUE;
				break;

			case NL_CONNECTIONLESS:

				// socket()
				osdBeginSignalRedoLoop();
				ip_handle = socket(PF_INET, SOCK_DGRAM, 0);
				osdEndSignalRedoLoop(INVALID_SOCKET_HANDLE == ip_handle);

				if (INVALID_SOCKET_HANDLE == ip_handle)
				{
					NL_PRINT(NL_OS, NL_WARN, "could not get socket handle");
					return_val =  0;
					break;
				}

				// bind()
				serv_addr.sin_family = PF_INET;
				serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
				serv_addr.sin_port = htons(inet_port[NL_LOCAL]);

				osdBeginSignalRedoLoop();
				r=bind(ip_handle, (PTR_SOCK_ADDR_STRUCT) &serv_addr,
						sizeof(serv_addr));
				osdEndSignalRedoLoop(0 != r);

				if (0 != r)
				{
					NL_PRINT(NL_OS, NL_WARN, "could not bind");
					nlCloseSocket(ip_handle);
					return_val = 0;
					break;
				}

				FindLocalPort();

				open = TRUE;
				return_val = 1;
				break;

			default:
				NL_PRINT(NL_APP, NL_ERROR, "connect type not set");
				break;
			}

			SetBlocking(blocking);
		}
		else
		{
			NL_PRINT(NL_APP, NL_ERROR, "invalid protocol");
			return_val =  0;
		}
	}
	else
	{
		NL_PRINT(NL_APP, NL_ERROR, "invalid domain");
		return_val = 0;
	}
	return return_val;
}

/******************************************************************************
void	NL_Channel::Close(void)

	close the socket

Keywords: NL_Channel, network
******************************************************************************/
void NL_Channel::Close(void)
{
	NL_PRINT(NL_NL, NL_LOG, "NL_Channel::Close");
	
	if (!open)
	{
		return;
	}

	if (domain & NL_INET)
	{
		if (protocol & NL_IP)
		{
			nlCloseSocket (ip_handle);
		}
		else
		{
			NL_PRINT(NL_APP, NL_ERROR, "invalid protocol");
		}
	}
	else
	{
		NL_PRINT(NL_APP, NL_ERROR, "invalid domain");
	}

	valid = FALSE;
	open = FALSE;
	q_specified = FALSE;
}

/******************************************************************************
NL_Channel	*NL_Channel::Accept(void)

	accept a connection (server)

Keywords: NL_Channel, network
******************************************************************************/
NL_Channel *NL_Channel::Accept(void)
{
	NL_PRINT(NL_NL, NL_LOG, "NL_Channel::Accept");

	NL_Channel *new_channel;
	INET_SOCK_ADDR_STRUCT cli_addr;
	memset((void *) &cli_addr, 0, sizeof(cli_addr));
	int clilen = sizeof(cli_addr);

	if (connection_type == NL_CONNECTIONLESS)
	{
		NL_PRINT(NL_APP, NL_ERROR, "cannot Accept with NL_CONNECTIONLESS channel");
	}
	
	if (!open)
	{
		NL_PRINT(NL_APP, NL_ERROR, "attempt to accept on non-open channel");
	}

	if (!q_specified)
	{
		if (!SpecifyQueue())
		{
			return NULL;
		}
	}

	new_channel = new NL_Channel();
	*new_channel = *this;

	if (domain & NL_INET)
	{
		if (protocol & NL_IP)
		{
			switch (connection_type)
			{
			case NL_CONNECTED:
				osdBeginSignalRedoLoop();
				new_channel->ip_handle =
					accept(ip_handle, (PTR_SOCK_ADDR_STRUCT) &cli_addr,
							&clilen);
				osdEndSignalRedoLoop(INVALID_SOCKET_HANDLE == new_channel->ip_handle);

				if (INVALID_SOCKET_HANDLE == new_channel->ip_handle)
				{
					if (OSD_EWOULDBLOCK == OSD_ERRNO)
					{
						last_errno = NL_CHANNEL_WOULD_BLOCK;
					}
					else
					{
						NL_PRINT(NL_OS, NL_WARN, "could not accept()");
					}
					delete new_channel;
					return NULL;
				}

				new_channel->inet_address[NL_REMOTE] = cli_addr.sin_addr.s_addr;
				new_channel->inet_port[NL_REMOTE] = ntohs(cli_addr.sin_port);
				break;

			default:
				NL_PRINT(NL_APP, NL_ERROR, "connect type not set");
				break;
			}
		}
		else
		{
			NL_PRINT(NL_APP, NL_ERROR, "invalid protocol");
			return 0;
		}
	}
	else
	{
		NL_PRINT(NL_APP, NL_ERROR, "invalid domain");
		return 0;
	}

	return new_channel;
}

/******************************************************************************
long	NL_Channel::SpecifyQueue(void)

	specify the server's willingness to accept incoming connections

Keywords: NL_Channel, network
******************************************************************************/
long NL_Channel::SpecifyQueue(void)
{
	NL_PRINT(NL_NL, NL_LOG, "NL_Channel::SpecifyQueue");

	long return_val;
	if (q_specified)
	{
		NL_PRINT(NL_APP, NL_ERROR, "queue already specified");
	}

	if (domain & NL_INET)
	{
		if (protocol & NL_IP)
		{
			int r;
			switch (connection_type)
			{
			case NL_CONNECTED:
				// listen ()
				osdBeginSignalRedoLoop();
				r = listen(ip_handle, 5);
				osdEndSignalRedoLoop(0 != r);

				if (0 != r)
				{
					NL_PRINT(NL_OS, NL_WARN, "could not listen");
					return_val =  0;
					break;
				}

				q_specified = TRUE;
				return_val = 1;
				break;

			case NL_CONNECTIONLESS:
				NL_PRINT(NL_APP, NL_ERROR, "do not specify queue for NL_CONNECTIONLESS");
				return_val =  0;
				break;

			default:
				NL_PRINT(NL_APP, NL_ERROR, "connect type not set");
				return_val =  0;
				break;
			}
		}
		else
		{
			NL_PRINT(NL_APP, NL_ERROR, "invalid protocol");
			return 0;
		}
	}
	else
	{
		NL_PRINT(NL_APP, NL_ERROR, "invalid domain");
		return 0;
	}

	return return_val;
}

/******************************************************************************
long	NL_Channel::Connect(void)

	connect client to server

Keywords: NL_Channel, network
******************************************************************************/
long NL_Channel::Connect(void)
{
	NL_PRINT(NL_NL, NL_LOG, "NL_Channel::Connect");

	if (domain & NL_INET)
	{
		if (protocol & NL_IP)
		{
			int r;
			INET_SOCK_ADDR_STRUCT serv_addr;
			switch (connection_type)
			{
			case NL_CONNECTED:
				memset((void *) &serv_addr, 0, sizeof(serv_addr));

				serv_addr.sin_family = PF_INET;
				serv_addr.sin_addr.s_addr = inet_address[NL_REMOTE];
				serv_addr.sin_port = htons(inet_port[NL_REMOTE]);

				osdBeginSignalRedoLoop();
				r = connect(ip_handle, (PTR_SOCK_ADDR_STRUCT) &serv_addr,
							sizeof(serv_addr));
				osdEndSignalRedoLoop(0 != r);

				if (0 != r)
				{
						if(osdGetLastError() == OSD_EISCONN)
						{ return TRUE;}
						return FALSE;
				}
				else
				{
						return TRUE;
				}

			case NL_CONNECTIONLESS:
				NL_PRINT(NL_APP, NL_ERROR,
						"attempt to connect NL_CONNECTIONLESS channel");
				break;

			default:
				NL_PRINT(NL_APP, NL_ERROR, "connect type not set");
				break;
			}

			return FALSE;
		}
		else
		{
			NL_PRINT(NL_APP, NL_ERROR, "invalid protocol");
			return FALSE;
		}
	}
	else
	{
		NL_PRINT(NL_APP, NL_ERROR, "invalid domain");
		return FALSE;
	}
}

/******************************************************************************
long	NL_Channel::Send(char *buffer, long nbytes)

	send data through socket

Keywords: NL_Channel, network
******************************************************************************/
long NL_Channel::Send(char *buffer, long nbytes)
{
	NL_PRINT(NL_NL, NL_LOG, "NL_Channel::Send");

	if (!open)
	{
		NL_PRINT(NL_APP, NL_ERROR, "channel not open");
		return -1;
	}

	if (domain & NL_INET)
	{
		if (protocol & NL_IP)
		{
			int r;
			switch (connection_type)
			{
			case NL_CONNECTED:
				osdBeginSignalRedoLoop();
				r = send(ip_handle, buffer, (int) nbytes, 0);
				osdEndSignalRedoLoop(-1 == r);

				if (-1 == r)
				{
					NL_PRINT(NL_OS, NL_ERROR, "send failed");
				}

				return r;

			case NL_CONNECTIONLESS:
				INET_SOCK_ADDR_STRUCT dest_addr;
				memset((void *) &dest_addr, 0, sizeof(dest_addr));

				dest_addr.sin_family = PF_INET;
				dest_addr.sin_addr.s_addr = inet_address[NL_REMOTE];
				dest_addr.sin_port = htons(inet_port[NL_REMOTE]);

				osdBeginSignalRedoLoop();
				r = sendto(ip_handle, buffer, (int) nbytes, 0,
							(PTR_SOCK_ADDR_STRUCT) &dest_addr, sizeof(dest_addr));
				osdEndSignalRedoLoop(-1 == r);

				if (-1 == r)
				{
					NL_PRINT(NL_OS, NL_ERROR, "send failed");
				}

				return r;

			default:
				NL_PRINT(NL_APP, NL_ERROR, "connect type not set");
				break;
			}

			return FALSE;
		}
		else
		{
			NL_PRINT(NL_APP, NL_ERROR, "invalid protocol");
			return FALSE;
		}
	}
	else
	{
		NL_PRINT(NL_APP, NL_ERROR, "invalid domain");
		return FALSE;
	}
}

/******************************************************************************
long	NL_Channel::Receive(char *buffer, long nbytes, long msec_timeout)

	recieve data through socket

Keywords: NL_Channel, network
******************************************************************************/
long NL_Channel::Receive(char *buffer, long nbytes, long msec_timeout)
{
	long tmp_nbytes;
	tmp_nbytes = nbytes;
	return Receive(buffer, &tmp_nbytes, msec_timeout);
}


/******************************************************************************
long	NL_Channel::Receive(char *buffer, long *nbytes, long msec_timeout)

	recieve data through socket

Keywords: NL_Channel, network
******************************************************************************/
long NL_Channel::Receive(char *buffer, long *nbytes, long msec_timeout)
{
	NL_PRINT(NL_NL, NL_LOG, "NL_Channel::Receive");

	if (!open)
	{
		NL_PRINT(NL_APP, NL_ERROR, "channel not open");
	}

	if (domain & NL_INET)
	{
		if (protocol & NL_IP)
		{
			char *buffer_ptr;
			int bytes_left;
			fd_set readfds;
			struct timeval timeout;
			int width;
			int n;
			int r;
			int fromlen = sizeof(INET_SOCK_ADDR_STRUCT);
			switch (connection_type)
			{
			case NL_CONNECTED:
				if (msec_timeout)
				{
					timeout.tv_sec = (int) (msec_timeout / 1000);
					timeout.tv_usec =
						1000 * (msec_timeout - (1000 * timeout.tv_sec));
					width = (int) (1 + ip_handle);
				}
				bytes_left = (int) (*nbytes);
				buffer_ptr = buffer;
				while (0 < bytes_left)
				{
					if (msec_timeout)
					{
						FD_ZERO(&readfds);
						FD_SET(ip_handle, &readfds);
#if OSD_OS==OSD_HPUX && OSD_OSVER==9
						n=select(width, (int *) &readfds,
							(int *) NULL, (int *) NULL, &timeout);
#else
						n=select(width, (fd_set *) &readfds,
							(fd_set *) NULL, (fd_set *) NULL, &timeout);
#endif
						if (0 > n)
						{
							NL_PRINT(NL_APP, NL_ERROR,
								"select failed in timeout receive");
							last_errno = NL_CHANNEL_ERROR;
							return NL_CHANNEL_ERROR;
						}
						else if (0 == n)
						{
							*nbytes -= bytes_left;
							last_errno = NL_CHANNEL_TIMED_OUT;
							return NL_CHANNEL_TIMED_OUT;
						}
						else if (!FD_ISSET(ip_handle, &readfds))
						{
							NL_PRINT(NL_APP, NL_ERROR,
								"select returned improperly");
							last_errno = NL_CHANNEL_ERROR;
							return NL_CHANNEL_ERROR;
						}
					}

					osdBeginSignalRedoLoop();
					r = recv(ip_handle, buffer_ptr, (int) bytes_left, 0);
					osdEndSignalRedoLoop(-1 == r);

					if (-1 == r)
					{
						if (OSD_EWOULDBLOCK == ERRNO)
						{
							last_errno = NL_CHANNEL_WOULD_BLOCK;
							return NL_CHANNEL_WOULD_BLOCK;
						}

						NL_PRINT(NL_OS, NL_WARN, "recv failed");
						last_errno = NL_CHANNEL_ERROR;
						return NL_CHANNEL_ERROR;
					}
					else if (0 == r)
					{
						break;
					}

					bytes_left -= r;
					buffer_ptr += r;
				}

				*nbytes -= bytes_left;
				return *nbytes;

			case NL_CONNECTIONLESS:
				INET_SOCK_ADDR_STRUCT orig_addr;
				memset((void *) &orig_addr, 0, sizeof(orig_addr));

				osdBeginSignalRedoLoop();
				r = recvfrom(ip_handle, buffer, (int) (*nbytes), 0,
							(PTR_SOCK_ADDR_STRUCT) &orig_addr, &fromlen);
				osdEndSignalRedoLoop(-1 == r);

				if (-1 == r)
				{
					if (OSD_EWOULDBLOCK == ERRNO)
					{
						last_errno = NL_CHANNEL_WOULD_BLOCK;
						return NL_CHANNEL_WOULD_BLOCK;
					}

					NL_PRINT(NL_OS, NL_ERROR, "could not receive");
				}

				inet_address[NL_RETURN] = orig_addr.sin_addr.s_addr;
				inet_port[NL_RETURN] = ntohs(orig_addr.sin_port);

				*nbytes = r;
				return r;

			default:
				NL_PRINT(NL_APP, NL_ERROR, "connect type not set");
				break;
			}

			return FALSE;
		}
		else
		{
			NL_PRINT(NL_APP, NL_ERROR, "invalid protocol");
			return FALSE;
		}
	}
	else
	{
		NL_PRINT(NL_APP, NL_ERROR, "invalid domain");
		return FALSE;
	}
}
