
/******************************************************************************

Coral Tree nl/message.cc
Copyright (C) 1998 Infinite Monkeys Incorporated

This program is free software; you can redistribute it and/or modify
it under the terms of the Artistic License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
Artistic License for more details.

You should have received a copy of the Artistic License along with
this program (see meta/ART_LICE); if not, write to coral@imonk.com.

******************************************************************************/

#include <nl/nl.h>

//TODO: remove the dependency on GPL_Message's protected (should be private)
//      data.  Currently, the InsertLong on the Send is copying the data
//      anyway, so there is no real performance gain by having such access.

/******************************************************************************

******************************************************************************/
NL_Message::NL_Message(void)
{
}

/******************************************************************************

******************************************************************************/
NL_Message::NL_Message(const NL_Message & msg)
{
	operator=(msg);
}

/******************************************************************************

******************************************************************************/
NL_Message::~NL_Message(void)
{
}

/******************************************************************************

******************************************************************************/
NL_Message & NL_Message::operator=(const NL_Message & msg)
{
	GPL_Message::operator=(msg);
	return *this;
}

/******************************************************************************

******************************************************************************/
GPL_Boolean NL_Message::Send(NL_Channel *channel)
{
	long r;

	
/******************************************************************************
Instead of using 2 sends, the length is inserted at the start of the buffer
of a single send.  2 recv()s are used to receive this single message.  The
reason for a single send is that using 2 sends take about an order of 
magnitude performance hit.  This is the case when TCP_NODELAY is not set.
******************************************************************************/
	InsertLong(length,0);

	r = channel->Send(buffer, length);
	if(r != length)
	{
		NL_PRINT(NL_OS,NL_WARN,"improper message send");
		return FALSE;
	}

	DeleteRange(0,3);
	return TRUE;
}

/******************************************************************************

******************************************************************************/
GPL_Boolean NL_Message::Receive(NL_Channel *channel, long msec_timeout)
{
	long n_length, r;

	Clear();
	r = channel->Receive((char *)&n_length, sizeof(long), msec_timeout);
	switch(r)
	{
		case 0:
			return FALSE;
		case NL_CHANNEL_WOULD_BLOCK:
			return NL_CHANNEL_WOULD_BLOCK;
		case NL_CHANNEL_ERROR:
			NL_PRINT(NL_OS,NL_WARN,"could not recieve message");
			return NL_CHANNEL_ERROR;
		case NL_CHANNEL_TIMED_OUT:
			return NL_CHANNEL_TIMED_OUT;
		default:
			if(r != sizeof(long))
			{
				NL_PRINT(NL_OS,NL_WARN,"recieved corrupt message");
			}
			break;
	}

	length = ntohl(n_length);
	CheckSize(length);

	// If this is a read-only or copied buffer of
	// insufficient length, make a separate buffer into
	// which the bytes can be read.  This keeps the
	// network clean and avoids polluting the heap.
	GPL_Array<char> temp;
	register char *buf = buffer;
	if (GPL_MESSAGE_READONLY & flags ||
		((GPL_MESSAGE_INVALID & flags) && (GPL_MESSAGE_COPY & flags)))
	{
		temp.Resize(length);
		buf = temp.GetData();
	}

	r = channel->Receive(buf, length, msec_timeout);
	switch(r)
	{
		case 0:
			return FALSE;
		case NL_CHANNEL_WOULD_BLOCK:
			return NL_CHANNEL_WOULD_BLOCK;
		case NL_CHANNEL_ERROR:
			NL_PRINT(NL_OS,NL_WARN,"could not recieve message");
			return NL_CHANNEL_ERROR;
		case NL_CHANNEL_TIMED_OUT:
			return NL_CHANNEL_TIMED_OUT;
		default:
			if(r != length)
			{
				NL_PRINT(NL_OS,NL_WARN,"recieved corrupt message");
			}
			break;
	}

	// If this is a read-only message...
	if (GPL_MESSAGE_READONLY & flags)
	{
		return FALSE;
	}
	
	// Copy the appropriate number of bytes out of the
	// temporary buffer into the message.  The caller
	// is responsible for checking GPL_MESSAGE_INVALID
	// to determine if the incoming message exceeded
	// the ability of the copied buffer to hold.
	if (buffer != buf)
	{
		osdMemcpy(buffer, buf, length);
	}

	return TRUE;
}


