#ifndef SPA_ERROR_H
#define SPA_ERROR_H

/******************************************************************************

Coral Tree spa/error.h
Copyright (C) 1998 Infinite Monkeys Incorporated

This program is free software; you can redistribute it and/or modify
it under the terms of the Artistic License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
Artistic License for more details.

You should have received a copy of the Artistic License along with
this program (see meta/ART_LICE); if not, write to coral@imonk.com.

******************************************************************************/

WBA_INIT_DECLARATION(SPA);

#define SPA_MAX_PRINT_TITLE_LEN		200
#define SPA_MAX_PRINT_STRING_LEN	1000

void spaPrintWithLocation(long source,long type,char *string,char *file,long line,long indent);

#define spaPrint(source,type,string)	spaPrintWithLocation(source,type,string,__FILE__,__LINE__,SPA_GLOBAL_INDENT)
#define spaIndent()						SPA_GLOBAL_INDENT+=4
#define spaUnIndent()					SPA_GLOBAL_INDENT-=4

/* SPA_PRINT sources */
#define SPA_SOFTWARE	0	/* software related condition */
#define SPA_SYSTEM		1	/* system related condition */
#define SPA_USER		2	/* improper usage by user of module */
#define SPA_GENERAL		3	/* non-specific SPA (specific is preferred) */

/* SPA_PRINT types */
#define SPA_LOG			0	/* verbose log updates */
#define SPA_WARN		1	/* possible problem */
#define SPA_ERROR		2	/* definite problem */
#define SPA_FATAL		3	/* fatal problem */

extern long SPA_GLOBAL_VERBOSITY;
extern long SPA_GLOBAL_STARTTIME;
extern long SPA_GLOBAL_LASTTIME;

void spaPrintf(char *format, ...);

#ifdef SPA_ENABLE_TIMECHECKING
#define spaTimeCheck(string) {												\
								long t, local;								\
								static long local_count = 0;				\
								static long local_total = 0;				\
								static long local_last = 0;					\
								static long local_steptotal = 0;			\
								t = osdTimer();								\
								if(!local_last) local_last = t;				\
								if(!SPA_GLOBAL_STARTTIME)					\
								{											\
									SPA_GLOBAL_STARTTIME = t;				\
									SPA_GLOBAL_LASTTIME = t;				\
								}											\
								local = t - local_last;						\
								local_steptotal += t - SPA_GLOBAL_LASTTIME;	\
								local_total += local;						\
								long av1, av2;								\
								if(local_count > 0)							\
								{											\
									av1 = local_steptotal / local_count;	\
									av2 = local_total	/ local_count;		\
								}											\
								else{av1 = 0;av2 = 0;}						\
								spaPrintf(									\
									"TIMER: G(%d %d) L(%d) A(%d %d) %s",	\
									t - SPA_GLOBAL_LASTTIME,				\
									t - SPA_GLOBAL_STARTTIME,				\
									local,									\
									av1,									\
									av2,									\
									string);								\
								local_count++;								\
								SPA_GLOBAL_LASTTIME = t;					\
								local_last = t;								\
							}
#else
#define spaTimeCheck(string)	/* NOOP */
#endif

#endif /* SPA_ERROR_H */
