#include <spa/spa.h>
#include <spa/lmdif1.h>

double ALPHA;
double TOL;

class I_Func : public SPA_Function1D
{
	private:
		double alpha, B, n;
		GPL_Boolean run;
		double lastRun;
		double tol;
	public:
		I_Func(double alpha_in, double B_in, double n_in, double t)
		{
			alpha = alpha_in;
			B = B_in;
			n = n_in;
			run = FALSE;
			tol = t;	
		}
virtual	~I_Func(){}

		operator double(void)
		{
			if(run)
			{
				return lastRun;
			}
			else
			{
				return Integrate();
			}
		}

		double Integrate()
		{
			run = TRUE;
			Set(SPA_MIDYDIFF, tol, 20, 5);
			lastRun = Rule(0.0, 1e100);
			return lastRun;
		}

		double Integrate_()
		{
			run = TRUE;
			Set(SPA_MIDEXP, tol, 20, 5);
			lastRun = Rule(0.0, 1e1);
			return lastRun;
		}

		double Function(double z)
		{
			if(B < 0.0)
			{
				wbaLog("SPA ERROR", "B is negative resulting in a complex number, try another seed");
			}
			return pow(z, (double)(n*(1.0-alpha)-1.0))*
				pow((1.0+B*z), (alpha-0.5))*exp(-z);
		}
};

class I_Func2 : public SPA_Function1D
{
	private:
		double alpha, B, n;
		GPL_Boolean run;
		double lastRun;
		double tol;
	public:
		I_Func2(double alpha_in, double B_in, double n_in, double t)
		{
			alpha = alpha_in;
			B = B_in;
			n = n_in;
			run = FALSE;
			tol = t;	
		}
virtual	~I_Func2(){}

		operator double(void)
		{
			if(run)
			{
				return lastRun;
			}
			else
			{
				return Integrate();
			}
		}

		double Integrate_()
		{
			run = TRUE;
			Set(SPA_MIDYDIFF, tol, 20, 5);
			lastRun = Rule(0.0, 1e100);
			return lastRun;
		}

		double Integrate()
		{
			run = TRUE;
			Set(SPA_MIDUNIT, tol, 20, 5);
			lastRun = Rule(0.0, 1e1);
			return lastRun;
		}

		double Function(double z)
		{
			if(B < 0.0)
			{
				wbaLog("SPA ERROR", "B is negative resulting in a complex number, try another seed");
			}
			return pow(z, (double)(n*(1.0-alpha)-1.0))*
				pow((1.0+B*z), (alpha-0.5))*exp(-z);
		}
};

class I_Func3 : public SPA_Function1D
{
	private:
		double alpha, B, n;
		GPL_Boolean run;
		double lastRun;
		double tol;
	public:
		I_Func3(double alpha_in, double B_in, double n_in, double t)
		{
			alpha = alpha_in;
			B = B_in;
			n = n_in;
			run = FALSE;
			tol = t;	
		}
virtual	~I_Func3(){}

		operator double(void)
		{
			if(run)
			{
				return lastRun;
			}
			else
			{
				return Integrate();
			}
		}

		double Integrate_()
		{
			run = TRUE;
			Set(SPA_MIDYDIFF, tol, 20, 5);
			lastRun = Rule(0.0, 1e100);
			return lastRun;
		}

		double Integrate()
		{
			run = TRUE;
			Set(SPA_MIDEXP, tol, 20, 5);
			lastRun = Rule(0.0, 1e1);
			return lastRun;
		}

		double Function(double z)
		{
			if(B < 0.0)
			{
				wbaLog("SPA ERROR", "B is negative resulting in a complex number, try another seed");
			}
			return pow(z, (double)(n*(1.0-alpha)-1.0))*
				pow((1.0+B*z), (alpha-0.5))*exp(-z);
		}
};

class I_Integrate : public SPA_Integrate
{
	public:
		double alpha, B, n;
		GPL_Boolean run;
		double lastRun;
		double tol;
	public:
		I_Integrate(double alpha_in, double B_in, double n_in, double t)
		{
			alpha = alpha_in;
			B = B_in;
			n = n_in;
			run = FALSE;
			tol = t;
		}
virtual	~I_Integrate(){}

		operator double(void)
		{
			if(run)
			{
				return lastRun;
			}
			else
			{
				return Integrate();
			}
		}

		void Set(double alpha_in)
		{
			alpha = alpha_in;
		}
		double Integrate()
		{
fprintf(stderr,"---------------\n");
			run = TRUE;
			double maxy = 1.0;
			double tuner = 0.1;
			double r1,r2;
			
//#define TRAP_MIN 1e-100
#define TRAP_MIN 0
#define TRAP_MAX 1e100
#if 0
			r1 = Trapezoidx(TRAP_MIN, TRAP_MAX, maxy);
			while(1)
			{
				maxy *= tuner;
				r2 = Trapezoidx(TRAP_MIN, TRAP_MAX, maxy);
fprintf(stderr, "%f Integrate old %.12f new %.12f\n", maxy, r1, r2);
if(maxy == 0.0) exit(-1);
				if(fabs(r2-r1) < tol)
				{
					lastRun = r2;
					break;
				}
				r1 = r2;
			}
#else
			r1 = Quadrature(TRAP_MIN, TRAP_MAX, maxy, 0, 0.0);
			while(1)
			{
				maxy *= tuner;
				r2 = Quadrature(TRAP_MIN, TRAP_MAX, maxy, 0, 0.0);
fprintf(stderr, "%f Integrate old %.12f new %.12f\n", maxy, r1, r2);
if(maxy == 0.0) exit(-1);
				if(fabs(r2-r1) < tol)
				{
					lastRun = r2;
					break;
				}
				r1 = r2;
			}
			//lastRun = Quadrature(TRAP_MIN, TRAP_MAX, tol, 0, 0.0);
#endif
			return lastRun;
		}
		double Function(double z)
		{
//fprintf(stderr,"n %f +++++++\n",n);
//fprintf(stderr,"%.12f\n", pow(z, (double)(n*(1.0-alpha)-1.0)));
//fprintf(stderr,"%.12f %.12f %.12f\n",(1.0+B*z), (alpha-0.5), pow((1.0+B*z), (alpha-0.5)));
//fprintf(stderr,"%.12f\n", exp(-z));
			if(B < 0.0)
			{
				wbaLog("SPA ERROR", "B is negative resulting in a complex number, try another seed");
			}
			return pow(z, (double)(n*(1.0-alpha)-1.0))*
				pow((1.0+B*z), (alpha-0.5))*exp(-z);
		}
};

double betasq(double alpha)
{
	//double tmp;
	double lg, g, g2;
	static double last_alpha = 0;
	static double last_betasq;
fprintf(stderr,"alpha %f\n", alpha);
	if(alpha == last_alpha) return last_betasq;
	lg = lgamma(1.0+alpha); 
	g = signgam*exp(lg);
	lg = lgamma(1.0+2.0*alpha); 
	g2 = signgam*exp(lg);
	//tmp = gamma(1.0+alpha);
	last_alpha = alpha;
	last_betasq = ((2.0*g*g)/(g2))-1.0;
fprintf(stderr,"Beta^2 %f\n", last_betasq);
	return last_betasq;
	//return ((2.0*tmp*tmp)/(gamma(1.0+2.0*alpha)))-1.0;
}

#define TOL_FCN 1e-5
double fcn(double alpha, double B)
{
	I_Integrate I1(alpha, B, 1, TOL_FCN);
	I_Integrate I2(alpha, B, 2, TOL_FCN);
	I_Integrate I3(alpha, B, 3, TOL_FCN);

	fprintf(stderr, "I3 %g\n", (double)I3);
//exit(0);
	fprintf(stderr, "I1 %f A %f y0 %f\n", (double)I1, 1.0/(double)I1, (double)I1/(double)I2);

	return ((double)I1*(double)I3) - ((double)1.0+(double)betasq(alpha))*(double)I2*(double)I2;
}

double fcn2(double alpha, double B)
{
	I_Func I1(alpha, B, 1, TOL_FCN);
	I_Func I2(alpha, B, 2, TOL_FCN);
	I_Func I3(alpha, B, 3, TOL_FCN);

	fprintf(stderr, "I3 %g\n", (double)I3);
//exit(0);
	fprintf(stderr, "I1 %f A %f y0 %f\n", (double)I1, 1.0/(double)I1, (double)I1/(double)I2);

	return ((double)I1*(double)I3) - ((double)1.0+(double)betasq(alpha))*(double)I2*(double)I2;
}

#undef TOL_FCN
#define TOL_FCN 1e-5
double fcn3(double alpha, double B)
{
	I_Func2 I1(alpha, B, 1, TOL_FCN);
	I_Func2 I2(alpha, B, 2, TOL_FCN);
	I_Func2 I3(alpha, B, 3, TOL_FCN);

	fprintf(stderr, "I3 %g\n", (double)I3);
//exit(0);
	fprintf(stderr, "I1 %f A %f y0 %f\n", (double)I1, 1.0/(double)I1, (double)I1/(double)I2);

	return ((double)I1*(double)I3) - ((double)1.0+(double)betasq(alpha))*(double)I2*(double)I2;
}

double fcn4(double alpha, double B)
{
	I_Func3 I1(alpha, B, 1, TOL_FCN);
	I_Func3 I2(alpha, B, 2, TOL_FCN);
	I_Func3 I3(alpha, B, 3, TOL_FCN);

	fprintf(stderr, "I3 %g\n", (double)I3);
//exit(0);
	fprintf(stderr, "I1 %f A %f y0 %f\n", (double)I1, 1.0/(double)I1, (double)I1/(double)I2);

	return ((double)I1*(double)I3) - ((double)1.0+(double)betasq(alpha))*(double)I2*(double)I2;
}

int lmdifFunction(int *m, int *n, double *x, double *fvec, int *iflag)
{
	double tol = TOL;
fprintf(stderr,"lmdifFunction %.12f %.12f %.12f\n", ALPHA, x[0], tol);
	I_Integrate I1(ALPHA, x[0], 1, tol);
	I_Integrate I2(ALPHA, x[0], 2, tol);
	I_Integrate I3(ALPHA, x[0], 3, tol);

	fvec[0] = (I1*I3) - (1.0+betasq(ALPHA))*I2*I2;

	return 0;
}

int lmdifFunction2(int *m, int *n, double *x, double *fvec, int *iflag)
{
	double tol = TOL;
fprintf(stderr,"lmdifFunction %.12f %.12f %.12f\n", ALPHA, x[0], tol);
	I_Func2 I1(ALPHA, x[0], 1, tol);
	I_Func2 I2(ALPHA, x[0], 2, tol);
	I_Func2 I3(ALPHA, x[0], 3, tol);

	fvec[0] = (I1*I3) - (1.0+betasq(ALPHA))*I2*I2;

	return 0;
}

int main(const int argc, const char *argv[], const char *envp[])
{
	//double alpha = 0.40;
	//double B = 9.03130;
	//B = 8.0;
	//printf("alpha %f B %f renders %f\n", alpha, B, fcn( alpha, B));

	{
		GPL_Aware gpl;
		SPA_Aware spa;
		WBA_ENV.LoadEnvVariables(envp, FALSE);
	}

	
	double x, fvec;
	double tol;
	int info = 0;
	int m = 1;
	int n = 1;
	int iwa;
	int lwa = m*n + 5*n +m;


#if 0
	tol = atof(argv[1]);
	TOL = tol;
	ALPHA = atof(argv[2]);
	x = pow(10.0,pow(ALPHA*2.0,2.0));
	double *wa = new double[lwa];
	lmdif1_(lmdifFunction2, &m, &n, &x, &fvec, &tol, &info, &iwa, wa, &lwa);
	fprintf(stderr,"1: info %d   B %f rendering %f\n", info, x, fvec);
#endif

#if 0
	double x, fvec;
	double tol = TOL;
	int info = 0;
	int m = 1;
	int n = 1;
	int iwa;
	int lwa = m*n + 5*n +m;
	x = 1.01388;
	//x = 13.01388;
	//x = 1000.0;
	//x = 1e-5;

	double *wa = new double[lwa];
	//fprintf(stderr,"func %f\n", fcn(0.60, 32.6538)); 
//exit(0);
	//lmdif1_(lmdifFunction, &m, &n, &x, &fvec, &tol, &info, &iwa, wa, &lwa);
	//fprintf(stderr,"0: info %d   B %f rendering %f\n", info, x, fvec);
	lmdif1_(lmdifFunction2, &m, &n, &x, &fvec, &tol, &info, &iwa, wa, &lwa);
	fprintf(stderr,"1: info %d   B %f rendering %f\n", info, x, fvec);

	exit(0);
#endif

#if 1
	double val;
	long tm;
	tm = osdTimer();
	val = fcn(0.40, 9.03);
	fprintf(stderr,"***** fnc %g %d\n", val, osdTimer() - tm);
	tm = osdTimer();
	val = fcn2(0.40, 9.03);
	fprintf(stderr,"***** YDIFF fnc2 %g %d\n", val, osdTimer() - tm);
	tm = osdTimer();
	val = fcn3(0.40, 9.03);
	fprintf(stderr,"***** UNIT fnc3 %g %d\n", val, osdTimer() - tm);
	tm = osdTimer();
	val = fcn4(0.40, 9.03);
	fprintf(stderr,"***** EXP fnc4 %g %d\n", val, osdTimer() - tm);
#endif
}

