#ifndef DNA_ERROR_H
#define DNA_ERROR_H

/******************************************************************************

Coral Tree dna/error.h
Copyright (C) 1998 Infinite Monkeys Incorporated

This program is free software; you can redistribute it and/or modify
it under the terms of the Artistic License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
Artistic License for more details.

You should have received a copy of the Artistic License along with
this program (see meta/ART_LICE); if not, write to coral@imonk.com.

******************************************************************************/

#include <gpl/gpl.h>

/***********************************************************************
class DNA_ErrorType

Network-layer error type.
***********************************************************************/
enum DNA_ErrorType
{
	// Bus
	DNA_ERROR_NONE					=	0x0000,
	DNA_ERROR_INSUFFICIENT_MEMORY	=	0x0001,
	DNA_ERROR_NO_DATA_AVAILABLE		=	0x0002,
	DNA_ERROR_BAD_HANDLE			=	0x0003,
	DNA_ERROR_BAD_MESSAGE			=	0x0004,
	DNA_ERROR_BAD_ADDRESS			=	0x0005,
	DNA_ERROR_INTERRUPTED_CALL		=	0x0006,
	DNA_ERROR_IO_ERROR				=	0x0007,
	DNA_ERROR_BAD_IMPLEMENTATION	=	0x0008,
	DNA_ERROR_BAD_STATE				=	0x0009,
	DNA_ERROR_UNKNOWN				=	0x000A,

	// Binding
	DNA_ERROR_INSUFFICIENT_RESOURCE	=	0x1001,
	DNA_ERROR_BAD_PROTOCOL			=	0x1002,
	DNA_ERROR_NO_PERMISSION			=	0x1003,
	DNA_ERROR_ADDRESS_IN_USE		=	0x1004,
	DNA_ERROR_ADDRESS_UNAVAILABLE	=	0x1005,

	// Connection
	DNA_ERROR_PROTOCOL_ERROR		=	0x2001,
	DNA_ERROR_BLOCK_IMPENDING		=	0x2002,
	DNA_ERROR_BAD_FAMILY			=	0x2003,
	DNA_ERROR_CONNECTION_IMPENDING	=	0x2004,
	DNA_ERROR_CONNECTION_REFUSED	=	0x2005,
	DNA_ERROR_CONNECTION_PENDING	=	0x2006,
	DNA_ERROR_HANDLE_CONNECTED		=	0x2007,
	DNA_ERROR_NETWORK_UNREACHABLE	=	0x2008,
	DNA_ERROR_TIME_OUT				=	0x2009,
	DNA_ERROR_CONNECTION_RESET		=	0x200A
};

extern const char *dnaErrorTypeToString(const DNA_ErrorType t);

/***********************************************************************
class DNA_Error

Network-layer error.
***********************************************************************/
class DNA_Error
{
	public:
		DNA_Error(const DNA_ErrorType t = DNA_ERROR_NONE, const int c = 0)
			: type(t), code(c) {}
		~DNA_Error(void) {}

		inline DNA_ErrorType GetType(void) const { return type; }
		inline int GetCode(void) const { return code; }

		void Set(const DNA_ErrorType t)
		{
			type = t;
			code = OSD_ERRNO;
			comment.Clear();
		}
		void Set(const DNA_ErrorType t, const int c)
		{
			type = t;
			code = c;
			comment.Clear();
		}
		void Set(const DNA_ErrorType t, const GPL_String & s)
		{
			type = t;
			code = OSD_ERRNO;
			comment = s;
		}
		void Set(const DNA_ErrorType t, const int c, const GPL_String & s)
		{
			type = t;
			code = c;
			comment = s;
		}

		GPL_String ToString(void) const;

	protected:
		DNA_ErrorType type;
		int code;
		GPL_String comment;
};

#endif /* DNA_ERROR_H */
