/******************************************************************************

Coral Tree dna/ipsocket.cc
Copyright (C) 1998 Infinite Monkeys Incorporated

This program is free software; you can redistribute it and/or modify
it under the terms of the Artistic License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
Artistic License for more details.

You should have received a copy of the Artistic License along with
this program (see meta/ART_LICE); if not, write to coral@imonk.com.

******************************************************************************/

#include <gpl/gpl.h>
#include <dna/dna.h>

/******************************************************************************
class DNA_IPSocket
******************************************************************************/
DNA_IPSocket::DNA_IPSocket(const int d, const int t)
	: domain (d), type(t), handle(INVALID_SOCKET_HANDLE)
{
}

DNA_IPSocket::DNA_IPSocket(const DNA_IPSocket & s)
	: domain(s.domain), type(s.type), handle(s.handle)
{
}

DNA_IPSocket::~DNA_IPSocket(void)
{
}

void DNA_IPSocket::SetBackLog(const unsigned long b)
{
	// listen ()
	int rc;
	osdBeginSignalRedoLoop();
	rc = listen(handle, (int) b);
	osdEndSignalRedoLoop(0 != rc);
}

void DNA_IPSocket::Reset(void)
{
	wbaLog("DNA LOG", "reset(%d)", handle);
	DNA_Error error;
	Disconnect(error);
	Close(error);
	handle = INVALID_SOCKET_HANDLE;
}

GPL_Boolean DNA_IPSocket::Accept(DNA_IPPoint *destination, DNA_IPSocket & s,
								DNA_Error & error)
{
	wbaLog("DNA LOG", "accept(%d, %s)", handle,
			destination->ToString().GetData());
	error.Set(DNA_ERROR_NONE, 0);

	// accept()
#if OSD_OS==OSD_LINUX
	unsigned int l = sizeof(destination->GetAddress());
#else
	int l = sizeof(destination->GetAddress());
#endif
	int rc;
	osdBeginSignalRedoLoop();
	rc = accept(handle, (PTR_IP_SOCKET_ADDRESS) &(destination->GetAddress()), &l);
	osdEndSignalRedoLoop(INVALID_SOCKET_HANDLE == rc);

	// Save the handle.
	s.handle = rc;

	if (INVALID_SOCKET_HANDLE == rc)
	{
		switch(OSD_ERRNO)
		{
		case OSD_EBADF:
		case OSD_ENOTSOCK:
			error.Set(DNA_ERROR_BAD_HANDLE);
			break;
		case OSD_EINTR:
			error.Set(DNA_ERROR_INTERRUPTED_CALL);
			break;
		case OSD_ENODEV:
		case OSD_EOPNOTSUPP:
		case OSD_EPROTO:
			error.Set(DNA_ERROR_BAD_PROTOCOL);
			break;
		case OSD_ENOMEM:
			error.Set(DNA_ERROR_INSUFFICIENT_MEMORY);
			break;
		case OSD_ENOSR:
			error.Set(DNA_ERROR_INSUFFICIENT_RESOURCE);
			break;
		case OSD_EWOULDBLOCK:
			error.Set(DNA_ERROR_NO_DATA_AVAILABLE);
			break;
		default:
			error.Set(DNA_ERROR_UNKNOWN);
			break;
		}
		wbaLog("DNA WARNING", "Accept: %s", error.ToString().GetData());
		return FALSE;
	}

	return TRUE;
}

GPL_Boolean DNA_IPSocket::Connect(DNA_IPPoint *destination, DNA_Error & error)
{
	wbaLog("DNA LOG", "connect(%d, %s)", handle,
			destination->ToString().GetData());
	error.Set(DNA_ERROR_NONE, 0);

	// connect()
	int rc;
	osdBeginSignalRedoLoop();
	rc = connect(handle,
			(PTR_IP_SOCKET_ADDRESS) &(((DNA_IPPoint *) destination)->GetAddress()),
				sizeof(((DNA_IPPoint *) destination)->GetAddress()));
	osdEndSignalRedoLoop(0 != rc);

	if (-1 == rc)
	{
		switch(OSD_ERRNO)
		{
		case OSD_EACCES:
			error.Set(DNA_ERROR_NO_PERMISSION);
			break;
		case OSD_EADDRINUSE:
			error.Set(DNA_ERROR_ADDRESS_IN_USE);
			break;
		case OSD_EADDRNOTAVAIL:
			error.Set(DNA_ERROR_ADDRESS_UNAVAILABLE);
			break;
		case OSD_EAFNOSUPPORT:
			error.Set(DNA_ERROR_BAD_PROTOCOL);
			break;
		case OSD_EBADF:
		case OSD_ENOENT:
		case OSD_ENOTSOCK:
			error.Set(DNA_ERROR_BAD_HANDLE);
			break;
		case OSD_ECONNREFUSED:
			error.Set(DNA_ERROR_CONNECTION_REFUSED);
			break;
		case OSD_EINPROGRESS:
			error.Set(DNA_ERROR_CONNECTION_PENDING);
			break;
		case OSD_EINTR:
			error.Set(DNA_ERROR_INTERRUPTED_CALL);
			break;
		case OSD_EINVAL:
			error.Set(DNA_ERROR_BAD_ADDRESS);
			break;
		case OSD_EIO:
		case OSD_ENXIO:
			error.Set(DNA_ERROR_IO_ERROR);
			break;
		case OSD_EISCONN:
			error.Set(DNA_ERROR_HANDLE_CONNECTED);
			break;
		case OSD_ENETUNREACH:
			error.Set(DNA_ERROR_NETWORK_UNREACHABLE);
			break;
		case OSD_ENOSR:
			error.Set(DNA_ERROR_INSUFFICIENT_RESOURCE);
			break;
		case OSD_ETIMEDOUT:
			error.Set(DNA_ERROR_TIME_OUT);
			break;
		default:
			error.Set(DNA_ERROR_UNKNOWN);
			break;
		}
		wbaLog("DNA WARNING", "Connect: %s", error.ToString().GetData());
		return FALSE;
	}

	return TRUE;
}

GPL_Boolean DNA_IPSocket::Disconnect(DNA_Error & error)
{
	wbaLog("DNA LOG", "disconnect(%d)", handle);
	error.Set(DNA_ERROR_NONE, 0);

	// NOOP
	return TRUE;
}

GPL_Boolean DNA_IPSocket::Open(DNA_Error & error)
{
	wbaLog("DNA LOG", "open()");
	error.Set(DNA_ERROR_NONE, 0);

	// socket()
	osdBeginSignalRedoLoop();
	handle = socket(domain, type, 0);
	osdEndSignalRedoLoop(INVALID_SOCKET_HANDLE == handle);

	if (INVALID_SOCKET_HANDLE == handle)
	{
		switch(OSD_ERRNO)
		{
		case OSD_EACCES:
			error.Set(DNA_ERROR_NO_PERMISSION);
			break;
		case OSD_EMFILE:
		case OSD_ENOSR:
			error.Set(DNA_ERROR_INSUFFICIENT_RESOURCE);
			break;
		case OSD_ENOMEM:
			error.Set(DNA_ERROR_INSUFFICIENT_MEMORY);
			break;
		case OSD_EPROTONOSUPPORT:
			error.Set(DNA_ERROR_BAD_PROTOCOL);
			break;
		default:
			error.Set(DNA_ERROR_UNKNOWN);
			break;
		}
		wbaLog("DNA WARNING", "Open: %s", error.ToString().GetData());
		return FALSE;
	}
	return TRUE;
}

GPL_Boolean DNA_IPSocket::Bind(DNA_Point *point, DNA_Error & error)
{
	wbaLog("DNA LOG", "bind(%d, %s)", handle, point->ToString().GetData());
	error.Set(DNA_ERROR_NONE, 0);

	if ((DNA_Point *) 0 == point)
	{
		error.Set(DNA_ERROR_BAD_ADDRESS, 0);
		return FALSE;
	}

	// bind()
	int rc;
	osdBeginSignalRedoLoop();
	rc = bind(handle, (PTR_IP_SOCKET_ADDRESS) &(((DNA_IPPoint *) point)->GetAddress()),
			sizeof(((DNA_IPPoint *) point)->GetAddress()));
	osdEndSignalRedoLoop(0 != rc);

	if (0 != rc)
	{
		switch(OSD_ERRNO)
		{
		case OSD_EACCES:
			error.Set(DNA_ERROR_NO_PERMISSION);
			break;
		case OSD_EADDRINUSE:
			error.Set(DNA_ERROR_ADDRESS_IN_USE);
			break;
		case OSD_EADDRNOTAVAIL:
			error.Set(DNA_ERROR_ADDRESS_UNAVAILABLE);
			break;
		case OSD_EBADF:
		case OSD_EINVAL:
		case OSD_ENOTSOCK:
			error.Set(DNA_ERROR_BAD_HANDLE);
			break;
		case OSD_ENOSR:
			error.Set(DNA_ERROR_INSUFFICIENT_RESOURCE);
			break;
		case OSD_EIO:
			error.Set(DNA_ERROR_IO_ERROR);
			break;
		default:
			// NOOP
			break;
		}
		wbaLog("DNA WARNING", "Bind: %s", error.ToString().GetData());
		Close(handle);
		return FALSE;
	}
	return TRUE;
}

GPL_Boolean DNA_IPSocket::Close(DNA_Error & error)
{
	wbaLog("DNA LOG", "close(%d)", handle);
	error.Set(DNA_ERROR_NONE, 0);

	Close(handle);
	return TRUE;
}

GPL_Boolean DNA_IPSocket::Close(SOCKET_HANDLE & handle)
{
	register GPL_Boolean success = TRUE;
#if OSD_OSTYPE==OSD_WIN32
	closesocket(handle);
#else /* OSD_OSTYPE!=OSD_WIN32 */
	success = success && (0 == shutdown(handle, 2));
	success = success && (0 == close(handle));
#endif /* OSD_OSTYPE!=OSD_WIN32 */
	handle = INVALID_SOCKET_HANDLE;
	return success;
}

GPL_Boolean DNA_IPSocket::ReadFrom(DNA_IPPoint *destination, void *buffer,
									const unsigned long size,
									unsigned long & length,
									DNA_Error & error)
{
	wbaLog("DNA LOG", "readfrom(%d, %s, %p, %d)", handle,
			destination->ToString().GetData(), buffer, size);
	error.Set(DNA_ERROR_NONE, 0);

	if ((DNA_Point *) 0 == destination)
	{
		error.Set(DNA_ERROR_BAD_ADDRESS, 0);
		return FALSE;
	}

	// Clear the destination address.
#if OSD_OS==OSD_LINUX
	unsigned int destinationLength = sizeof(((DNA_IPPoint *) destination)->GetAddress());
#else
	int destinationLength = sizeof(((DNA_IPPoint *) destination)->GetAddress());
#endif
	osdMemset((void *) &(((DNA_IPPoint *) destination)->GetAddress()), 0,
				destinationLength);

	// recvfrom()
	int rc;
	osdBeginSignalRedoLoop();
	rc = recvfrom(handle, (char *) buffer, (int) size, 0,
					(PTR_IP_SOCKET_ADDRESS) &(((DNA_IPPoint *) destination)->GetAddress()),
					&destinationLength);
	osdEndSignalRedoLoop(-1 == rc);

	if (-1 == rc)
	{
		switch(OSD_ERRNO)
		{
		case OSD_EBADF:
		case OSD_ENOTSOCK:
		case OSD_ESTALE:
			error.Set(DNA_ERROR_BAD_HANDLE);
			break;
		case OSD_EINTR:
			error.Set(DNA_ERROR_INTERRUPTED_CALL);
			break;
		case OSD_EIO:
			error.Set(DNA_ERROR_IO_ERROR);
			break;
		case OSD_ENOMEM:
			error.Set(DNA_ERROR_INSUFFICIENT_MEMORY);
			break;
		case OSD_ENOSR:
			error.Set(DNA_ERROR_INSUFFICIENT_RESOURCE);
			break;
		case OSD_EWOULDBLOCK:
			error.Set(DNA_ERROR_NO_DATA_AVAILABLE);
			break;
		default:
			error.Set(DNA_ERROR_UNKNOWN);
			break;
		}
		wbaLog("DNA WARNING", "ReadFrom: %s", error.ToString().GetData());
		return FALSE;
	}

	length = rc;
	return TRUE;
}

GPL_Boolean DNA_IPSocket::ReadFrom(DNA_IPPoint *destination, void *buffer,
									const unsigned long size,
									const long timeout, unsigned long & length,
									DNA_Error & error)
{
//!!TODO: Actually implement a timeout.
	long t;
	t = timeout;
return DNA_IPSocket::ReadFrom(destination, buffer, size, length, error);
}

GPL_Boolean DNA_IPSocket::Read(void *buffer, const unsigned long size,
								unsigned long & length, DNA_Error & error)
{
	return DNA_IPSocket::Read(buffer, size, DNA_BLOCK_INDEFINITELY, length, error);
}

GPL_Boolean DNA_IPSocket::Read(void *buffer, const unsigned long size,
								const long timeout, unsigned long & length,
								DNA_Error & error)
{
	wbaLog("DNA LOG", "read(%d, %p, %d, %d)", handle, buffer, size, timeout);
	error.Set(DNA_ERROR_NONE, 0);

	// Initialize the timeout.
	struct timeval to;
	struct timeval *top = (struct timeval *) 0;
	if (DNA_MULTIPLEXOR_BLOCK != timeout)
	{
		to.tv_sec = (int) (timeout / 1000);
		to.tv_usec = 1000 * (timeout - (1000 * to.tv_sec));
		top = &to;
	}

	int rc;
	unsigned long bytes_left = size;
	length = 0;
	char *buffer_ptr;
	buffer_ptr = (char *)buffer;

	while(0 < bytes_left)
	{

		if(DNA_MULTIPLEXOR_BLOCK != timeout)
		{

			// Prepare to read.
			fd_set readfds;
			FD_ZERO(&readfds);
			FD_SET(handle, &readfds);

			// Check for outstanding messages.
			rc = DNA_SELECT(1 + handle, &readfds, NULL, NULL, top);
			if (0 > rc)
			{
				wbaLog("DNA WARNING", "select() failed");
				error.Set(DNA_ERROR_IO_ERROR);
				return FALSE;
			}

			if (0 == rc)
			{
				error.Set(DNA_ERROR_NO_DATA_AVAILABLE);
				return FALSE;
			}

			if (!FD_ISSET(handle, &readfds))
			{
				wbaLog("DNA WARNING", "select() returned improperly");
				error.Set(DNA_ERROR_UNKNOWN);
				return FALSE;
			}
		}

		osdBeginSignalRedoLoop();
		rc = recv(handle, (char *) buffer_ptr, (int) size, 0);
		osdEndSignalRedoLoop(-1 == rc);

		if (0 == rc)
		{
			error.Set(DNA_ERROR_NO_DATA_AVAILABLE);
			return FALSE;
		}

		if (-1 == rc)
		{
			switch(OSD_ERRNO)
			{
			case OSD_EBADF:
			case OSD_ENOTSOCK:
			case OSD_ESTALE:
				error.Set(DNA_ERROR_BAD_HANDLE);
				break;
			case OSD_EINTR:
				error.Set(DNA_ERROR_INTERRUPTED_CALL);
				break;
			case OSD_EIO:
				error.Set(DNA_ERROR_IO_ERROR);
				break;
			case OSD_ENOMEM:
				error.Set(DNA_ERROR_INSUFFICIENT_MEMORY);
				break;
			case OSD_ENOSR:
				error.Set(DNA_ERROR_INSUFFICIENT_RESOURCE);
				break;
			case OSD_EWOULDBLOCK:
				error.Set(DNA_ERROR_NO_DATA_AVAILABLE);
				break;
			case OSD_ECONNRESET:
				error.Set(DNA_ERROR_CONNECTION_RESET);
				break;
			default:
				error.Set(DNA_ERROR_UNKNOWN);
				break;
			}
			wbaLog("DNA WARNING", "Read: %s", error.ToString().GetData());
			return FALSE;
		}

		length += rc;
		buffer_ptr += rc;
		bytes_left -= rc;
	}
	return TRUE;
}

GPL_Boolean DNA_IPSocket::WriteTo(DNA_IPPoint *destination, void *buffer,
									const unsigned long size,
									unsigned long & length, DNA_Error & error)
{
	wbaLog("DNA LOG", "writeto(%d, %s, %p, %d)", handle,
			destination->ToString().GetData(), buffer, size);
	error.Set(DNA_ERROR_NONE, 0);

	if ((DNA_IPPoint *) 0 == destination)
	{
		error.Set(DNA_ERROR_BAD_ADDRESS, 0);
		return FALSE;
	}

	// sendto()
	int rc;
	osdBeginSignalRedoLoop();
	rc = sendto(handle, (char *) buffer, (int) size, 0,
				(PTR_IP_SOCKET_ADDRESS) &(((DNA_IPPoint *) destination)->GetAddress()),
				sizeof(((DNA_IPPoint *) destination)->GetAddress()));
	osdEndSignalRedoLoop(-1 == rc);

	if (-1 == rc)
	{
		// EWOULDBLOCK is blatantly ignored.
		switch(OSD_ERRNO)
		{
		case OSD_EBADF:
		case OSD_ENOTSOCK:
			error.Set(DNA_ERROR_BAD_HANDLE);
			break;
		case OSD_EINTR:
			error.Set(DNA_ERROR_INTERRUPTED_CALL);
			break;
		case OSD_EINVAL:
			error.Set(DNA_ERROR_BAD_ADDRESS);
			break;
		case OSD_EMSGSIZE:
		case OSD_ENOSR:
			error.Set(DNA_ERROR_INSUFFICIENT_RESOURCE);
			break;
		case OSD_ENOMEM:
			error.Set(DNA_ERROR_INSUFFICIENT_MEMORY);
			break;
		default:
			// NOOP
			break;
		}
		wbaLog("DNA WARNING", "WriteTo: %s", error.ToString().GetData());
		return FALSE;
	}

	length = rc;
	return TRUE;
}

GPL_Boolean DNA_IPSocket::Write(void *buffer, const unsigned long size,
						unsigned long & length, DNA_Error & error)
{
	wbaLog("DNA LOG", "write(%d, %p, %d)", handle, buffer, size);
	error.Set(DNA_ERROR_NONE, 0);

	int rc;
	osdBeginSignalRedoLoop();
	rc = send(handle, (char *) buffer, (int) size, 0);
	osdEndSignalRedoLoop(-1 == rc);

	if (-1 == rc)
	{
		switch(OSD_ERRNO)
		{
		case OSD_EBADF:
		case OSD_ENOTSOCK:
			error.Set(DNA_ERROR_BAD_HANDLE);
			break;
		case OSD_EINTR:
			error.Set(DNA_ERROR_INTERRUPTED_CALL);
			break;
		case OSD_EINVAL:
			error.Set(DNA_ERROR_BAD_ADDRESS);
			break;
		case OSD_ENOMEM:
			error.Set(DNA_ERROR_INSUFFICIENT_MEMORY);
			break;
		case OSD_EMSGSIZE:
		case OSD_ENOSR:
			error.Set(DNA_ERROR_INSUFFICIENT_RESOURCE);
			break;
		case OSD_EWOULDBLOCK:
			error.Set(DNA_ERROR_NO_DATA_AVAILABLE);
			break;
		default:
			error.Set(DNA_ERROR_UNKNOWN);
			break;
		}
		wbaLog("DNA WARNING", "Write: %s", error.ToString().GetData());
		return FALSE;
	}

	length = rc;
	return TRUE;
}

GPL_Boolean DNA_IPSocket::SetOptions(const int level, const int option,
										const void *buffer, const int length)
{
	wbaLog("DNA LOG", "setsockopt(%d, %d, %d, %p, %d)", handle,
			level, option, buffer, length);
	//!! TODO: Determine if setsockopt() is supported
	//!! by the Evil Empire.
	return 0 == setsockopt(handle, level, option, (const char *) buffer, length);
}

int DNA_IPSocket::Ioctl(long cmd, void *arg)
{
	wbaLog("DNA LOG", "ioctl(%d, %d, %p)", handle, cmd, arg);
#if	OSD_OSTYPE==OSD_WIN32
	unsigned long argument = *((char *)arg);
	return ioctlsocket(handle, cmd, &argument);
#else	/* OSD_OSTYPE!=OSD_WIN32 */
	return ioctl(handle, (int) cmd, (char *) arg);
#endif	/* OSD_OSTYPE!=OSD_WIN32 */
}

OSD_SignalHandler(nl_IPSocketSIGPIPE_Handler)
{
	wbaLog("DNA WARNING", "Caught broken pipe");
}

void DNA_IPSocket::GlobalInitialize(void)
{
#if OSD_OSTYPE==OSD_WIN32
	WORD wVersionRequested = MAKEWORD(DNA_HIGH_BYTE, DNA_LOW_BYTE);
	WSADATA wsaData;
	int err = WSAStartup(wVersionRequested, &wsaData);

	if (0 != err)
	{
		fprintf(stderr, "Could not find usable winsock.dll\n");
		wbaLog("DNA ERROR", "Winsock startup failed, could not find usable winsock.dll");
	}

	if (DNA_LOW_BYTE != LOBYTE(wsaData.wVersion) ||
		DNA_HIGH_BYTE != HIBYTE(wsaData.wVersion))
	{
		WSACleanup();
		wbaLog("DNA ERROR", "Winsock startup failed, could not find usable winsock.dll");
	}
#else /* OSD_OSTYPE!=OSD_WIN32 */
	OSD_SetSignalHandler(nl_IPSocketSIGPIPE_Handler, SIGPIPE);
#endif /* OSD_OSTYPE!=OSD_WIN32 */
}

void DNA_IPSocket::GlobalFinalize(void)
{
#if OSD_OSTYPE==OSD_WIN32
	int err = WSACleanup();
	if (err)
	{
		wbaLog("DNA WARNING", "Could not clean up sockets");
	}
#else /* OSD_OSTYPE!=OSD_WIN32 */
	// NOP.
#endif /* OSD_OSTYPE==OSD_WIN32 */
}

GPL_OBJECT_IMPLEMENTATION(DNA_IPSocket, 0x40070001, "DNA_IPSocket");
